'''Module for the legacy cut plane detection.

As implemented in:
https://bbpcode.epfl.ch/source/xref/platform/BlueRepairSDK/BlueRepairSDK/src/repair.cpp#263
'''
import logging
from collections import defaultdict

import numpy as np
from morph_tool import apical_point_section_segment
from neurom import iter_sections
from neurom.core import Section
from neurom.core.dataformat import COLS

from neuror.utils import RepairType, repair_type_map

L = logging.getLogger(__name__)


def children_ids(section):
    '''
    https://bbpcode.epfl.ch/source/xref/platform/BlueRepairSDK/BlueRepairSDK/src/helper_dendrite.cpp#111

    The original code returns the ids of the descendant sections
    but this implementation return the Section objects instead.
    '''
    return list(section.ipreorder())


def cut_detect(neuron, cut, offset, axis):
    '''Detect the cut leaves the old way

    The cut leaves are simply the leaves that live
    on the half-space (split along the 'axis' coordinate)
    with the biggest number of leaves
    '''
    count_plus = count_minus = sum_plus = sum_minus = 0

    for leaf in iter_sections(neuron, iterator_type=Section.ileaf):
        coord = leaf.points[-1, axis]
        if coord > offset:
            count_plus += 1
            sum_plus += coord
        else:
            count_minus += 1
            sum_minus += coord

    if count_plus == 0 or count_minus == 0:
        raise Exception("cut detection warning:one of the sides is empty. can't decide on cut side")

    if -sum_minus / count_minus > sum_plus / count_plus:
        sign = 1
    else:
        sign = -1

    for leaf in iter_sections(neuron, iterator_type=Section.ileaf):
        if leaf.points[-1, axis] * sign > offset:
            cut[leaf] = True

    return sign


def internal_cut_detection(neuron, axis):
    '''As in:

    https://bbpcode.epfl.ch/source/xref/platform/BlueRepairSDK/BlueRepairSDK/src/repair.cpp#263

    Use cut_detect to get the side of the half space the points live in.
    Then mark points which are children of the apical section.

'''
    axis = {'x': COLS.X, 'y': COLS.Y, 'z': COLS.Z}[axis.lower()]

    cut = defaultdict(lambda key: False)
    side = cut_detect(neuron, cut, 0, axis)

    # reclassify cut points in tuft,based on apical point position
    apical_section_id, point_id = apical_point_section_segment(neuron)
    if apical_section_id is not None:
        apical_section = neuron.sections[apical_section_id]
        apical_offset = apical_section.points[point_id, axis]
        cut_mark(children_ids(apical_section), cut, apical_offset, side, axis)
    else:
        apical_section = None

    extended_types = repair_type_map(neuron, apical_section)
    oblique_roots = get_obliques(neuron, extended_types)

    # reclassify points in obliques. based on the position of their root.
    for root in oblique_roots:
        offset = root.points[0]

        # FIXME: z is hard coded in the original code as well. It's probably a bug.
        cut_mark(root.children, cut, offset[COLS.Z], side, axis)

    cut_leaves = np.array([sec.points[-1, COLS.XYZ] for sec, is_cut in cut.items() if is_cut])

    return cut_leaves, side


def get_obliques(neuron, extended_types):
    '''
    Returns the oblique roots.

    https://bbpcode.epfl.ch/source/xref/platform/BlueRepairSDK/BlueRepairSDK/src/helper_dendrite.cpp#212
    '''
    return [section for section in iter_sections(neuron)
            if (extended_types[section] == RepairType.oblique and
                (section.parent is None or extended_types[section.parent] == RepairType.trunk))]


def cut_mark(sections, cut, offset, side, axis):
    '''
    https://bbpcode.epfl.ch/source/xref/platform/BlueRepairSDK/BlueRepairSDK/src/helper_dendrite.cpp#654
    '''
    for sec in sections:
        if sec.children:
            cut[sec] = False
            continue

        r = sec.points[-1, axis]

        growing_back = False
        mysec = sec

        while mysec.parent:
            for point in mysec.points:
                mr = point[axis]
                growing_back |= (mr - (r + (float(side) * 15.0))) * side > 0
            if growing_back:
                break
            mysec = mysec.parent

        cut[sec] = (r - offset) * side > 0 and not growing_back
    return cut



'''Tests in this file are to be run only with the [plotly] extra installed'''
from pathlib import Path
from tempfile import TemporaryDirectory

from click.testing import CliRunner
from numpy.testing import assert_equal

from neuror.cli import cli

DATA = Path(__file__).parent / 'data'


def test_report():
    with TemporaryDirectory('test-report') as folder:
        runner = CliRunner()
        result = runner.invoke(cli, ['report', str(DATA), folder])
        assert_equal(result.exit_code, 0)



'''Module for the detection of the cut plane'''
import json
import logging
import itertools as it
import operator
from pathlib import Path
import warnings
from typing import List, Union

import neurom as nm
import numpy as np

from neurom import geom
from neurom.core import Section, Neuron
from scipy import optimize, special

from neuror.cut_plane import legacy_detection, planes

L = logging.getLogger(__name__)


class CutPlane(planes.HalfSpace):
    '''The cut plane class

    It is composed of a HalfSpace and a morphology
    The morphology is part of the HalfSpace, the cut space is
    the complementary HalfSpace
    '''

    def __init__(self, coefs: List[float],
                 upward: bool,
                 morphology: Union[None, str, Path, Neuron],
                 bin_width: float):
        '''Cut plane ctor.

        Args:
            coefs: the [abcd] coefficients of a plane equation: a X + b Y + c Z + d = 0
            upward: if true, the morphology points satisfy: a X + b Y + c Z + d > 0
                    else, they satisfy: a X + b Y + c Z + d < 0
            morphology: the morphology
            bin_width: the bin width
        '''
        super().__init__(coefs[0], coefs[1], coefs[2], coefs[3], upward)

        if isinstance(morphology, Neuron):
            self.morphology = morphology
        elif isinstance(morphology, (str, Path)):
            self.morphology = nm.load_neuron(morphology)
        elif morphology is not None:
            raise Exception(f'Unsupported morphology type: {type(morphology)}')

        self.bin_width = bin_width
        self.cut_leaves_coordinates = None
        self.status = None
        self.minus_log_prob = None
        if morphology is not None:
            self._compute_cut_leaves_coordinates()
            self._compute_probabilities()
            self._compute_status()

    @classmethod
    def from_json(cls, cut_plane_obj, morphology=None):
        '''Factory constructor from a JSON file

        cut_plane_obj (dict|str|pathlib.Path): a cut plane
            It can be a python dictionary or a path to a json file that contains one
        '''
        assert isinstance(cut_plane_obj, (str, dict, Path))

        if not isinstance(cut_plane_obj, dict):
            with open(cut_plane_obj) as f:
                cut_plane_obj = json.load(f)

        data = cut_plane_obj['cut-plane']
        obj = CutPlane([data['a'], data['b'], data['c'], data['d']],
                       data['upward'],
                       morphology,
                       cut_plane_obj['details']['bin-width'])

        obj.cut_leaves_coordinates = np.array(cut_plane_obj['cut-leaves'])
        obj.status = cut_plane_obj['status']
        obj.minus_log_prob = cut_plane_obj['details']['-LogP']
        return obj

    # pylint: disable=arguments-differ
    @classmethod
    def from_rotations_translations(cls, transformations, morphology, bin_width):
        plane = planes.PlaneEquation.from_rotations_translations(transformations)
        return cls(plane.coefs, True, morphology, bin_width)

    @classmethod
    def find(cls, neuron, bin_width=3,
             searched_axes=('X', 'Y', 'Z'),
             searched_half_spaces=(-1, 1),
             fix_position=None):
        """
        Find and return the cut plane that is oriented along X, Y or Z.
        6 potential positions are considered: for each axis, they correspond
        to the coordinate of the first and last point of the neuron.

        Description of the algorithm:

        1) The distribution of all points along X, Y and Z is computed
           and put into 3 histograms.
        2) For each histogram we look at the first and last empty bins
           (ie. the last bin before the histogram starts rising,
           and the first after it reaches zero again). Under the assumption
           that there is no cut plane, the posteriori probability
           of observing this empty bin given the value of the not-empty
           neighbour bin is then computed.
        3) The lowest probability of the 6 probabilities (2 for each axes)
           corresponds to the cut plane

        Args:
            neuron (Neuron|str|pathlib.Path): a Neuron object or path
            bin_width: The size of the binning

            display: where or not to display the control plots
                     Note: It is the user responsability to call matplotlib.pyplot.show()

            searched_axes: x, y or z. Specify the planes for which to search the cut plane

            searched_half_spaces: A negative value means the morphology lives
                on the negative side of the plane, and a positive one the opposite.

            fix_position: If not None, this is the position
                for which to search the cut plane. Only the orientation will be searched for.
                This can be useful to find the orientation of a plane whose position is known.

        Returns:
            A cut plane object
        """
        warnings.warn("This method will be deprecated in favor or cut_leaves.find_cut_leaves"
                      "it also has bugs if one uses +1 in searched_half_spaces and multiple"
                      "combinations of searched_arguments.", DeprecationWarning)
        if not isinstance(neuron, Neuron):
            neuron = nm.load_neuron(neuron)

        # pylint: disable=invalid-unary-operand-type
        coef_d = -fix_position if fix_position is not None else 0

        cut_planes = [CutPlane([int(axis.upper() == 'X'), int(axis.upper() == 'Y'),
                                int(axis.upper() == 'Z'), coef_d],
                               upward=(side > 0),
                               morphology=neuron,
                               bin_width=bin_width)
                      for axis, side in it.product(searched_axes, searched_half_spaces)]

        best_cut_plane = max(cut_planes, key=operator.attrgetter('minus_log_prob'))

        if fix_position is None:
            _, bins = best_cut_plane.histogram()
            # The orientation of the plane is defined such as the morphology lives
            # on the positive side once coordinates are projected
            # (see HalfSpace.project_on_directed_normal)
            # So the first bin is where we look for the cut plane
            best_cut_plane.coefs[3] = bins[0]
        else:
            best_cut_plane.coefs[3] = coef_d
        return CutPlane(
            best_cut_plane.coefs, best_cut_plane.upward, neuron, best_cut_plane.bin_width)

    @classmethod
    def find_legacy(cls, neuron, axis):
        '''Find the cut points according to the legacy algorithm

        As implemented in:
        https://bbpcode.epfl.ch/source/xref/platform/BlueRepairSDK/BlueRepairSDK/src/repair.cpp#263
        '''
        if not isinstance(neuron, Neuron):
            neuron = nm.load_neuron(neuron)

        cut_leaves, side = legacy_detection.internal_cut_detection(neuron, axis)

        plane = cls([int(axis.upper() == 'X'), int(axis.upper() == 'Y'),
                     int(axis.upper() == 'Z'), 0],
                    upward=(side < 0),
                    morphology=None,
                    bin_width=0)
        plane.morphology = nm.load_neuron(neuron)
        plane.cut_leaves_coordinates = cut_leaves
        return plane

    @property
    def cut_sections(self):
        '''Returns sections that ends within the cut plane'''
        leaves = np.array([leaf
                           for neurite in self.morphology.neurites
                           for leaf in nm.iter_sections(neurite, iterator_type=Section.ileaf)])
        leaves_coord = [leaf.points[-1, nm.COLS.XYZ] for leaf in leaves]
        return leaves[self.distance(leaves_coord) < self.bin_width]

    def _compute_cut_leaves_coordinates(self):
        '''Returns cut leaves coordinates'''
        leaves = [leaf.points[-1, nm.COLS.XYZ] for leaf in self.cut_sections]
        self.cut_leaves_coordinates = np.vstack(leaves) if leaves else np.array([])

    def to_json(self):
        '''Return a dictionary with the following items:

            - status: 'ok' if everything went right, else an informative string
            - cut_plane: a tuple (plane, position) where 'plane' is 'X', 'Y' or 'Z'
                       and 'position' is the position
            - cut_leaves: an np.array of all termination points in the cut plane
            - figures: if 'display' option was used, a dict where values are tuples (fig, ax)
                     for each figure
            - details: A dict currently only containing -LogP of the bin where the cut plane was
                     found
        '''
        return {'cut-leaves': self.cut_leaves_coordinates,
                'status': self.status,
                'details': {'-LogP': self.minus_log_prob, 'bin-width': self.bin_width},
                'cut-plane': super().to_json()}

    def histogram(self):
        '''Get the point distribution projected along the normal to the plane

        Returns:
            a numpy.histogram
        '''
        points = _get_points(self.morphology)
        projected_points = self.project_on_directed_normal(points)
        min_, max_ = np.min(projected_points, axis=0), np.max(projected_points, axis=0)
        binning = np.arange(min_, max_ + self.bin_width, self.bin_width)
        return np.histogram(projected_points, bins=binning)

    def _compute_probabilities(self):
        '''Returns -log(p) where p is the a posteriori probabilities of the observed values
        in the bins X min, X max, Y min, Y max, Z min, Z max

        Parameters:
            hist: a dict of the X, Y and Z 1D histograms

        Returns: a dict of -log(p) values'''

        hist = self.histogram()
        if not hist[0].size:
            self.minus_log_prob = np.nan
        else:
            self.minus_log_prob = get_minus_log_p(0, hist[0][0])
        self._compute_status()
        return self.minus_log_prob

    def _compute_status(self):
        '''Returns ok if the probability that there is a cut plane is high enough'''
        _THRESHOLD = 50
        if np.isnan(self.minus_log_prob):
            self.status = 'The proba is NaN, something went wrong'
        elif self.minus_log_prob < _THRESHOLD:
            self.status = ('The probability that there is in fact NO '
                           'cut plane is high: -log(p) = {0} !').format(self.minus_log_prob)
        else:
            self.status = 'ok'


def _success_function(params, points, bin_width):
    '''The success function is low (=good) when the difference of points
    on the left side and right side of the plane is high'''
    plane = planes.PlaneEquation.from_rotations_translations(params)
    n_left, n_right = plane.count_near_plane(points, bin_width)
    res = -abs(n_left - n_right)
    return res


def _minimize(x0, points, bin_width):
    '''Returns a tuple of the optimized values of
    (rot_x, rot_y, rot_z, transl_x, transl_y, transl_z)'''
    delta_angle = 10  # in degrees
    delta_transl = 10
    delta = np.array([delta_angle, delta_angle, delta_angle,
                      delta_transl, delta_transl, delta_transl])
    bounds_min = x0 - delta
    bounds_max = x0 + delta
    bounds = list(zip(bounds_min, bounds_max))
    result = optimize.minimize(_success_function,
                               x0=x0,
                               args=(points, bin_width),
                               bounds=bounds,
                               method='Nelder-Mead')

    if result.status:
        raise Exception(result.message)
    return result.x


def get_minus_log_p(k, mu):
    '''Compute -Log(p|k) where p is the a posteriori probability to observe k counts
    in bin given than the mean value was "mu":
    demo: p(k|mu) = exp(-mu) * mu**k / k!
    '''
    return mu - k * np.log(mu) + np.log(special.factorial(k))


def _get_points(neuron):
    return np.array([point
                     for neurite in (neuron.neurites or [])
                     for section in nm.iter_sections(neurite)
                     for point in section.points])


def plot(neuron, result, inline=False):
    '''Plot the neuron, the cut plane and the cut leaves.

    Args:
        neuron (Neuron): the neuron to be plotted
        result (dict): the cut plane object in dictionary form
        inline (bool): if True, plot as an interactive plot (for example in a Jupyter notebook)
    '''
    try:
        from plotly_helper.neuron_viewer import NeuronBuilder
        from plotly_helper.object_creator import scatter
        from plotly_helper.shapes import line
    except ImportError as e:
        raise ImportError(
            'neuror[plotly] is not installed.'
            ' Please install it by doing: pip install neuror[plotly]') from e

    bbox = geom.bounding_box(neuron)

    plane = result['cut-plane']

    for display_plane, idx in [('xz', 0), ('yz', 1), ('3d', None)]:
        builder = NeuronBuilder(neuron, display_plane, inline=inline,
                                line_width=4, title='{}'.format(neuron.name))
        if idx is not None:
            if plane['a'] == 0 and plane['b'] == 0:
                builder.helper.add_shapes([
                    line(bbox[0][idx], -plane['d'], bbox[1][idx], -plane['d'], width=4)
                ])
            builder.helper.add_data({'a': scatter(result['cut-leaves'][:, [idx, 2]],
                                                  showlegend=False, width=5)})
        else:
            builder.helper.add_data({'a': scatter(result['cut-leaves'], width=2)})
        builder.plot()



'''Dendritic repair module

It is based on the BlueRepairSDK's implementation
'''
import logging
from collections import Counter, OrderedDict, defaultdict
from enum import Enum
from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, Optional

import morphio
import neurom as nm
import numpy as np
from morph_tool import apical_point_section_segment
from morph_tool.spatial import point_to_section_segment
from morphio import PointLevel, SectionType
from neurom import NeuriteType, iter_neurites, iter_sections, load_neuron
from neurom.core.dataformat import COLS
from neurom.core import Section
from neurom.features.sectionfunc import branch_order, section_path_length
from nptyping import NDArray
from scipy.spatial.distance import cdist

from neuror import axon, cut_plane
from neuror.utils import RepairType, direction, repair_type_map, rotation_matrix, section_length

_PARAMS = {
    'seg_length': 5.0,  # lenghts of new segments
    'sholl_layer_size': 10,  # resoluion of the shll profile
    'noise_continuation': 0.5,  # together with seg_length, this controls the tortuosity
    'soma_repulsion': 0.2,  # if 0, previous section direction, if 1, radial direction
    'bifurcation_angle': 20,  # noise amplitude in degree around mean bif angle on the cell
    'path_length_ratio': 0.5,  # a smaller value will make a strornger taper rate
    'children_diameter_ratio': 0.8,  # 1: child diam = parent diam, 0: child diam = tip diam
    'tip_percentile': 25,  # percentile of tip radius distributions to use as tip radius
}

# Epsilon can not be to small otherwise leaves stored in json files
# are not found in the NeuroM neuron
EPSILON = 1e-6

L = logging.getLogger('neuror')


class Action(Enum):
    '''To bifurcate or not to bifurcate ?'''
    BIFURCATION = 1
    CONTINUATION = 2
    TERMINATION = 3


def is_cut_section(section, cut_points):
    '''Return true if the section is close from the cut plane'''
    if cut_points.size == 0 or section.points.size == 0:
        return False
    return np.min(cdist(section.points[:, COLS.XYZ], cut_points)) < EPSILON


def is_branch_intact(branch, cut_points):
    '''Does the branch have leaves belonging to the cut plane ?'''
    return all(not is_cut_section(section, cut_points) for section in branch.ipreorder())


def _get_sholl_layer(section, origin, sholl_layer_size):
    '''Returns this section sholl layer'''
    return int(np.linalg.norm(section.points[-1, COLS.XYZ] - origin) / sholl_layer_size)


def _get_sholl_proba(sholl_data, section_type, sholl_layer, pseudo_order):
    '''Return the probabilities of bifurcation, termination and bifurcation
    in a dictionnary for the given sholl layer and branch order.

    If no data are available for this branch order, the action_counts
    are averaged on all branch orders for this sholl layer


    Args:
        sholl_data: nested dict that stores the number of section per SectionType, sholl order
            sholl layer and Action type
            sholl_data[neurite_type][layer][order][action_type] = counts
        section_type (SectionType): section type
        sholl_layer (int): sholl layer
        pseudo_order (int): pseudo order

    Returns:
        Dict[Action, float]: probability of each action

    Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n398  # noqa, pylint: disable=line-too-long
    Note2: OrderedDict ensures the reproducibility of np.random.choice outcome
    '''

    section_type_data = sholl_data[section_type]
    try:
        data_layer = section_type_data[sholl_layer]
    except KeyError:
        return OrderedDict([(Action.BIFURCATION, 0),
                            (Action.CONTINUATION, 0),
                            (Action.TERMINATION, 1)])

    try:
        action_counts = data_layer[pseudo_order]
    except KeyError:
        # No data for this order. Average on all orders for this layer
        # As done in https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n426  # noqa, pylint: disable=line-too-long

        action_counts = Counter()
        for data in data_layer.values():
            for action, count in data.items():
                action_counts[action] += count

    total_counts = sum(action_counts.values())
    if total_counts == 0:
        return OrderedDict([(Action.BIFURCATION, 0),
                            (Action.CONTINUATION, 0),
                            (Action.TERMINATION, 1)])

    res = OrderedDict([tuple((action, action_counts[action] / float(total_counts))) for action in
                       sorted(action_counts.keys(), key=lambda t: t.value)])
    return res


def _grow_until_sholl_sphere(
    section, origin, sholl_layer, params, taper, tip_radius, current_trunk_radius
):
    '''Grow until reaching next sholl layer

    Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n363  # noqa, pylint: disable=line-too-long
    '''
    backwards_sections = 0
    while _get_sholl_layer(
        section, origin, params['sholl_layer_size']
    ) == sholl_layer and backwards_sections < 4:
        _continuation(section, origin, params, taper(current_trunk_radius), tip_radius)

        # make sure we don't grow back the origin
        is_last_segment_toward_origin = np.dot(
            section.points[-1, COLS.XYZ] - origin,
            _last_segment_vector(section)) < 0

        if is_last_segment_toward_origin:
            backwards_sections += 1
    return backwards_sections


def _last_segment_vector(section, normalized=False):
    '''Returns the vector formed by the last 2 points of the section'''
    vec = section.points[-1, COLS.XYZ] - section.points[-2, COLS.XYZ]
    if normalized:
        return vec / np.linalg.norm(vec)
    return vec


def _branching_angles(section, order_offset=0):
    '''Return a list of 2-tuples. The first element is the branching order and the second one is
    the angles between the direction of the section and its children's ones

    Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/morphstats.cpp#n194  # noqa, pylint: disable=line-too-long
    '''
    if section_length(section) < EPSILON:
        return []
    res = []

    branching_order = branch_order(section) - order_offset
    for child in section.children:
        if section_length(child) < EPSILON:
            continue

        theta = np.math.acos(np.dot(direction(section), direction(child)) /
                             (section_length(section) * section_length(child)))
        res.append((branching_order, theta))
    return res


def _continuation(sec, origin, params, taper, tip_radius):
    '''Continue growing the section

    Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n241  # noqa, pylint: disable=line-too-long
    '''
    # The following lines is from BlueRepairSDK's code but I'm not
    # convinced by its relevance
    # if first_sec:
    #     direction_ = sec.points[-1, COLS.XYZ] - sec.points[0, COLS.XYZ]
    # else:
    section_direction = _last_segment_vector(sec)
    section_direction /= np.linalg.norm(section_direction)

    # better to use last section point, to get direction at the tip from where we grow
    radial_direction = sec.points[-1, COLS.XYZ] - origin
    length = np.linalg.norm(radial_direction)
    if length > EPSILON:
        radial_direction /= length
    else:
        radial_direction = sec.points[1, COLS.XYZ] - sec.points[0, COLS.XYZ]
        radial_direction /= np.linalg.norm(radial_direction)

    # NOTE: This is not an equiprobability uniform generator
    # It is not drawing a point inside a sphere but a cube.
    # so the probability of drawing in direction of the corners is higher
    noise_direction = (2 * np.random.random(size=3) - 1)
    noise_direction /= np.linalg.norm(noise_direction)

    direction_ = (
        section_direction
        + params["soma_repulsion"] * radial_direction
        + params["noise_continuation"] * noise_direction
    )
    direction_ /= np.linalg.norm(direction_)

    coord = sec.points[-1, COLS.XYZ] + direction_ * params["seg_length"]
    # we apply tapering function + minimum diameter
    radius = max(tip_radius, sec.points[-1, COLS.R] - taper)
    new_point = np.append(coord, radius)
    sec.points = np.vstack((sec.points, new_point))


def _y_cylindrical_extent(section):
    '''Returns the distance from the last section point to the origin in the XZ plane'''
    xz_last_point = section.points[-1, [0, 2]]
    return np.linalg.norm(xz_last_point)


def _max_y_dendritic_cylindrical_extent(neuron):
    '''Return the maximum distance of dendritic section ends and the origin in the XZ plane'''
    return max((_y_cylindrical_extent(section) for section in neuron.iter()
                if section.type in {SectionType.basal_dendrite, SectionType.apical_dendrite}),
               default=0)


class Repair(object):
    '''The repair class'''

    def __init__(self,  # pylint: disable=too-many-arguments
                 inputfile: Path,
                 axons: Optional[Path] = None,
                 seed: Optional[int] = 0,
                 cut_leaves_coordinates: Optional[NDArray[(3, Any)]] = None,
                 legacy_detection: bool = False,
                 repair_flags: Optional[Dict[RepairType, bool]] = None,
                 apical_point: NDArray[3, float] = None,
                 params: Dict = None):
        '''Repair the input morphology

        The repair algorithm uses sholl analysis of intact branches to grow new branches from cut
        leaves. The algorithm is fairly complex, but can be controled via a few parameters in the
        params dictionary. By default, they are:
        _PARAMS = {
            'seg_length': 5.0,  # lenghts of new segments
            'sholl_layer_size': 10,  # resoluion of the shll profile
            'noise_continuation': 0.5,  # together with seg_length, this controls the tortuosity
            'soma_repulsion': 0.2,  # if 0, previous section direction, if 1, radial direction
            'bifurcation_angle': 20,  # noise amplitude in degree around mean bif angle on the cell
            'path_length_ratio': 0.5,  # a smaller value will make a strornger taper rate
            'children_diameter_ratio': 0.8,  # 1: child diam = parent diam, 0: child diam = tip diam
            'tip_percentile': 25,  # percentile of tip radius distributions to use as tip radius
        }

        Args:
            inputfile: the input neuron to repair
            axons: donor axons whose section will be used to repair this axon
            seed: the numpy seed
            cut_leaves_coordinates: List of 3D coordinates from which to start the repair
            legacy_detection: if True, use the legacy cut plane detection
                (see neuror.legacy_detection)
            repair_flags: a dict of flags where key is a RepairType and value is whether
                it should be repaired or not. If not provided, all types will be repaired.
            apical_point: 3d vector for apical point, else, the automatic apical detection is used
                if apical_point == -1, no automatic detection will be tried
            params: repair internal parameters (see comments in code for details)

        Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/repair.cpp#n469  # noqa, pylint: disable=line-too-long
        '''
        np.random.seed(seed)
        self.legacy_detection = legacy_detection
        self.inputfile = inputfile
        self.axon_donors = axons or list()
        self.donated_intact_axon_sections = list()
        self.repair_flags = repair_flags or dict()
        self.params = params if params is not None else _PARAMS

        CutPlane = cut_plane.CutPlane
        if legacy_detection:
            self.cut_leaves = CutPlane.find_legacy(inputfile, 'z').cut_leaves_coordinates
        elif cut_leaves_coordinates is None:
            self.cut_leaves = CutPlane.find(inputfile, bin_width=15).cut_leaves_coordinates
        else:
            self.cut_leaves = np.asarray(cut_leaves_coordinates)

        self.neuron = load_neuron(inputfile)
        self.repair_type_map = dict()
        self.max_y_cylindrical_extent = _max_y_dendritic_cylindrical_extent(self.neuron)
        self.max_y_extent = max(np.max(section.points[:, COLS.Y])
                                for section in self.neuron.iter())

        self.info = dict()
        apical_section_id = None
        if apical_point != -1:
            if apical_point:
                # recall MorphIO ID = NeuroM ID - 1
                apical_section_id = point_to_section_segment(self.neuron, apical_point)[0] - 1
            else:
                apical_section_id, _ = apical_point_section_segment(self.neuron)

        if apical_section_id:
            self.apical_section = self.neuron.sections[apical_section_id]
        else:
            self.apical_section = None

        # record the tip radius as a lower bound for diameters with taper, excluding axons
        # because they are treated separately, with thinner diameters
        _diameters = [
            np.mean(leaf.points[:, COLS.R])
            for neurite in self.neuron.neurites
            if neurite.type is not NeuriteType.axon
            for leaf in iter_sections(neurite, iterator_type=Section.ileaf)
        ]
        self.tip_radius = (
            np.percentile(_diameters, self.params["tip_percentile"]) if _diameters else None
        )
        self.current_trunk_radius = None

        # estimate a global tapering rate of the morphology as a function of trunk radius,
        # such that the tip radius is attained on average af max_path_length, defined as a fraction
        # of the maximal path length via the parameter path_lengt_ratio. The smaller this parameter
        # the faster the radii will convert to tip_radius.
        # TODO: maybe we would need this per neurite_type
        max_path_length = self.params["path_length_ratio"] * np.max(
            nm.get("terminal_path_lengths_per_neurite", self.neuron)
        )
        self.taper = (
            lambda trunk_radius: (trunk_radius - self.tip_radius)
            * self.params["seg_length"]
            / max_path_length
        )

    # pylint: disable=too-many-locals, too-many-branches
    def run(self,
            outputfile: Path,
            plot_file: Optional[Path] = None):
        '''Run'''
        if self.cut_leaves.size == 0:
            L.warning('No cut leaves. Nothing to repair for morphology %s', self.inputfile)
            self.neuron.write(outputfile)
            return

        # See https://github.com/BlueBrain/MorphIO/issues/161
        keep_axons_alive = list()

        for axon_donor in self.axon_donors:
            if self.legacy_detection:
                plane = cut_plane.CutPlane.find_legacy(axon_donor, 'z')
            else:
                plane = cut_plane.CutPlane.find(axon_donor)
            keep_axons_alive.append(plane)
            self.donated_intact_axon_sections.extend(
                [section for section in iter_sections(plane.morphology)
                 if section.type == SectionType.axon and
                 is_branch_intact(section, plane.cut_leaves_coordinates)])

        self._fill_repair_type_map()
        self._fill_statistics_for_intact_subtrees()
        intact_axonal_sections = [section for section in iter_sections(self.neuron)
                                  if section.type == SectionType.axon and
                                  is_branch_intact(section, self.cut_leaves)]

        # BlueRepairSDK used to have a bounding cylinder filter but
        # I don't know what is it good at so I have commented
        # the only relevant line
        # bounding_cylinder_radius = 10000
        cut_sections_in_bounding_cylinder = [
            section for section in iter_sections(self.neuron)
            if (is_cut_section(section, cut_points=self.cut_leaves)
                # and np.linalg.norm(section.points[-1, COLS.XZ]) < bounding_cylinder_radius
                )
        ]

        used_axon_branches = set()

        cut_leaves_ids = {section: len(section.points)
                          for section in cut_sections_in_bounding_cylinder}

        for section in sorted(cut_sections_in_bounding_cylinder, key=section_path_length):
            type_ = self.repair_type_map[section]
            if not self.repair_flags.get(type_, True):
                L.debug('Repair flag set to False for section type %s : --> Skipping repair.',
                        type_)
                continue
            L.debug('Repairing: %s, section id: %s', type_, section.id)
            if type_ in {RepairType.basal, RepairType.oblique, RepairType.tuft}:
                self.current_trunk_radius = [
                    sec.points[0, COLS.R] for sec in section.iupstream()
                ][-1]
                origin = self._get_origin(section)
                if section.type == NeuriteType.basal_dendrite:
                    _continuation(section, origin, self.params,
                                  self.taper(self.current_trunk_radius), self.tip_radius)
                self._grow(section, self._get_order_offset(section), origin)
            elif type_ == RepairType.axon:
                axon.repair(self.neuron, section, intact_axonal_sections,
                            self.donated_intact_axon_sections, used_axon_branches,
                            self.max_y_extent)
            elif type_ == RepairType.trunk:
                L.debug('Trunk repair is not (nor has ever been) implemented')
            else:
                raise Exception('Unknown type: {}'.format(type_))

        if plot_file is not None:
            try:
                from neuror.view import plot_repaired_neuron
                plot_repaired_neuron(self.neuron, cut_leaves_ids, plot_file)
            except ImportError:
                L.warning('Skipping writing plots as [plotly] extra is not installed')

        self.neuron.write(outputfile)
        L.debug('Repair successful for %s', self.inputfile)

    def _find_intact_obliques(self):
        '''
        Find root sections of all intact obliques

        Root obliques are obliques with a section parent of type 'trunk'

        Note: based on
        https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n193
        '''
        root_obliques = (section for section in iter_sections(self.neuron)
                         if (self.repair_type_map[section] == RepairType.oblique and
                             not section.is_root() and
                             self.repair_type_map[section.parent] == RepairType.trunk))
        intacts = [oblique for oblique in root_obliques
                   if is_branch_intact(oblique, self.cut_leaves)]
        return intacts

    def _find_intact_sub_trees(self):
        '''Returns intact neurites

        There is a fallback mechanism in case there are no intact basals:
        https://bbpcode.epfl.ch/source/xref/platform/BlueRepairSDK/BlueRepairSDK/src/repair.cpp#658
        '''
        basals = [neurite.root_node for neurite in iter_neurites(self.neuron)
                  if (neurite.type == NeuriteType.basal_dendrite and
                      is_branch_intact(neurite.root_node, self.cut_leaves))]

        if not basals:
            L.debug("No intact basals found. Falling back on less strict selection.")
            basals = [section for section in iter_sections(self.neuron)
                      if (section.type == NeuriteType.basal_dendrite and
                          not is_cut_section(section, self.cut_leaves))]

        axons = [neurite.root_node for neurite in iter_neurites(self.neuron)
                 if (neurite.type == NeuriteType.axon and
                     is_branch_intact(neurite.root_node, self.cut_leaves))]
        obliques = self._find_intact_obliques()

        tufts = [section for section in iter_sections(self.neuron)
                 if (self.repair_type_map[section] == RepairType.tuft and
                     not is_cut_section(section, self.cut_leaves))]

        return basals + obliques + axons + tufts

    def _intact_branching_angles(self, branches):
        '''
        Returns lists of branching angles stored in a nested dict
        1st key: section type, 2nd key: branching order

        Args:
            branches (List[Neurite])

        Returns:
            Dict[SectionType, Dict[int, List[int]]]: Branching angles
        '''
        res = defaultdict(lambda: defaultdict(list))
        for branch in branches:
            order_offset = self._get_order_offset(branch)
            for section in branch.ipreorder():
                for order, angle in _branching_angles(section, order_offset):
                    res[self.repair_type_map[section]][order].append(angle)
        return dict(res)

    def _best_case_angle_data(self, section_type, branching_order):
        '''Get the distribution of branching angles for this section type and
        branching order

        If no data are available, fallback on aggregate data

        ..note:: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n329  # noqa, pylint: disable=line-too-long
        '''
        angles = self.info['intact_branching_angles'][section_type]
        return angles[branching_order] or list(chain.from_iterable(angles.values()))

    def _get_origin(self, branch):
        '''Return what should be considered as the origin for this branch'''
        if self.repair_type_map[branch] == RepairType.oblique:
            return branch.points[0, COLS.XYZ]
        if self.repair_type_map[branch] == RepairType.tuft:
            return self.apical_section.points[-1, COLS.XYZ]
        return self.neuron.soma.center

    def _get_order_offset(self, branch):
        r'''
        Return what should be considered as the branch order offset for this branch

        For obliques, the branch order is computed with respect to
        the branch order of the first oblique section so we have to remove the offset.

                    3
                 2  |
     oblique ---->\ |
                   \| 1
                    |
                    |
                    | 0

        oblique order is 2 - 1 = 1
        '''
        if self.repair_type_map[branch] == RepairType.oblique:
            return branch_order(branch)
        if self.repair_type_map[branch] == RepairType.tuft:
            return branch_order(self.apical_section)
        return 0

    def _compute_sholl_data(self, branches):
        '''Compute the number of termination, bifurcation and continuation section for each
        neurite type, sholl layer and shell order

        data[neurite_type][layer][order][action_type] = counts

        Args:
            branches: a collection of Neurite or Section that will be traversed

        Note: This is based on
        https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/morphstats.cpp#n93
        '''
        data = defaultdict(lambda: defaultdict(dict))

        for branch in branches:
            origin = self._get_origin(branch)
            order_offset = self._get_order_offset(branch)

            for section in branch.ipreorder():
                repair_type = self.repair_type_map[section]
                assert repair_type == self.repair_type_map[branch], \
                    'RepairType should not change along the branch way'
                order = branch_order(section) - order_offset
                first_layer, last_layer = (
                    np.linalg.norm(section.points[[0, -1], COLS.XYZ] - origin, axis=1)
                    // self.params['sholl_layer_size']
                ).astype(int)
                per_type = data[repair_type]

                starting_layer = min(first_layer, last_layer)

                # TODO: why starting_layer + 1 and not starting_layer ?
                # bcoste The continuation from the starting layer should be taken into account
                # But that's how it is done in:
                # https://bbpcode.epfl.ch/source/xref/platform/BlueRepairSDK/BlueRepairSDK/src/morphstats.cpp#88
                for layer in range(starting_layer + 1, max(first_layer, last_layer)):
                    if order not in per_type[layer]:
                        per_type[layer][order] = {Action.TERMINATION: 0,
                                                  Action.CONTINUATION: 0,
                                                  Action.BIFURCATION: 0}
                    per_type[layer][order][Action.CONTINUATION] += 1

                if order not in per_type[last_layer]:
                    per_type[last_layer][order] = {Action.TERMINATION: 0,
                                                   Action.CONTINUATION: 0,
                                                   Action.BIFURCATION: 0}

                per_type[last_layer][order][Action.BIFURCATION if section.children else
                                            Action.TERMINATION] += 1
        return data

    def _bifurcation(self, section, order_offset):
        '''Create 2 children at the end of the current section

        Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n287  # noqa, pylint: disable=line-too-long
        '''
        current_diameter = section.points[-1, COLS.R] * 2

        child_diameters = 2 * [self.params["children_diameter_ratio"] * current_diameter]

        median_angle = np.median(
            self._best_case_angle_data(
                self.repair_type_map[section], branch_order(section) - order_offset
            )
        )

        last_segment_vec = _last_segment_vector(section)
        orthogonal = np.cross(last_segment_vec, section.points[-1, COLS.XYZ])

        def shuffle_direction(direction_):
            '''Return the direction to which to grow a child section'''
            theta = median_angle + np.radians(np.random.random() * self.params['bifurcation_angle'])
            first_rotation = np.dot(rotation_matrix(orthogonal, theta), direction_)
            new_dir = np.dot(rotation_matrix(direction_, np.random.random() * np.pi * 2),
                             first_rotation)
            return new_dir / np.linalg.norm(new_dir)

        for child_diameter in child_diameters:
            child_start = section.points[-1, COLS.XYZ]
            points = [child_start.tolist(),
                      (child_start + shuffle_direction(last_segment_vec) *
                       self.params['seg_length']).tolist()]
            prop = PointLevel(points, [current_diameter, child_diameter])

            child = section.append_section(prop)
            L.debug('section appended: %s', child.id)

    def _grow(self, section, order_offset, origin):
        '''grow main method

        Will either:
            - continue growing the section
            - create a bifurcation
            - terminate the growth

        Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.cpp#n387  # noqa, pylint: disable=line-too-long
        '''
        if (self.repair_type_map[section] == RepairType.tuft and
                _y_cylindrical_extent(section) > self.max_y_cylindrical_extent):
            return

        sholl_layer = _get_sholl_layer(section, origin, self.params['sholl_layer_size'])
        pseudo_order = branch_order(section) - order_offset
        L.debug('In _grow. Layer: %s, order: %s', sholl_layer, pseudo_order)

        proba = _get_sholl_proba(self.info['sholl'],
                                 self.repair_type_map[section],
                                 sholl_layer,
                                 pseudo_order)

        L.debug('action proba[%s][%s][%s]: %s', section.type, sholl_layer, pseudo_order, proba)
        action = np.random.choice(list(proba.keys()), p=list(proba.values()))

        if action == Action.CONTINUATION:
            L.debug('Continuing')
            backwards_sections = _grow_until_sholl_sphere(section, origin, sholl_layer,
                                                          self.params, self.taper,
                                                          self.tip_radius,
                                                          self.current_trunk_radius)

            if backwards_sections == 0:
                self._grow(section, order_offset, origin)
            L.debug(section.points[-1])

        elif action == Action.BIFURCATION:
            L.debug('Bifurcating')
            backwards_sections = _grow_until_sholl_sphere(section, origin, sholl_layer,
                                                          self.params, self.taper,
                                                          self.tip_radius,
                                                          self.current_trunk_radius)
            if backwards_sections == 0:
                self._bifurcation(section, order_offset)
                for child in section.children:
                    self.repair_type_map[child] = self.repair_type_map[section]
                    self._grow(child, order_offset, origin)

    def _fill_statistics_for_intact_subtrees(self):
        '''Compute statistics'''
        branches = self._find_intact_sub_trees()

        self.info = dict(
            intact_branching_angles=self._intact_branching_angles(branches),
            dendritic_sections=[section for section in iter_sections(self.neuron)
                                if section.type in {nm.APICAL_DENDRITE, nm.BASAL_DENDRITE}],
            sholl=self._compute_sholl_data(branches),
        )

    def _fill_repair_type_map(self):
        '''Assign a repair section type to each section

        Note: based on https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/repair.cpp#n242  # noqa, pylint: disable=line-too-long
        '''
        self.repair_type_map = repair_type_map(self.neuron, self.apical_section)


def repair(inputfile: Path,  # pylint: disable=too-many-arguments
           outputfile: Path,
           axons: Optional[List[Path]] = None,
           seed: int = 0,
           cut_leaves_coordinates: Optional[NDArray[(3, Any)]] = None,
           legacy_detection: bool = False,
           plot_file: Optional[Path] = None,
           repair_flags: Optional[Dict[RepairType, bool]] = None,
           apical_point: List = None,
           params: Dict = None):
    '''The repair function

    Args:
        inputfile: the input morph
        outputfile: the output morph
        axons: the axons
        seed: the numpy seed
        cut_leaves_coordinates: List of 3D coordinates from which to start the repair
        plot_file: the filename of the plot
        repair_flags: a dict of flags where key is a RepairType and value is whether
            it should be repaired or not. If not provided, all types will be repaired.
        apical_point: 3d vector for apical point, else, the automatic apical detection is used
        params: repair internal parameters, None will use defaults
    '''
    ignored_warnings = (
        # We append the section at the wrong place and then we reposition them
        # afterwards so we can ignore the warning
        morphio.Warning.wrong_duplicate,

        # the repair process creates new sections that may not have siblings
        morphio.Warning.only_child,

        # We are appending empty section and filling them later on
        morphio.Warning.appending_empty_section,
    )

    for warning in ignored_warnings:
        morphio.set_ignored_warning(warning, True)

    if axons is None:
        axons = list()

    if params is None:
        params = _PARAMS

    obj = Repair(inputfile, axons=axons, seed=seed, cut_leaves_coordinates=cut_leaves_coordinates,
                 legacy_detection=legacy_detection, repair_flags=repair_flags,
                 apical_point=apical_point, params=params)
    obj.run(outputfile, plot_file=plot_file)

    for warning in ignored_warnings:
        morphio.set_ignored_warning(warning, False)



"""Detect cut leaves with new algo."""
from itertools import product
import numpy as np
from neurom.core.dataformat import COLS
from neuror.cut_plane.planes import HalfSpace


def _get_cut_leaves(half_space, morphology, bin_width, percentile_threshold):
    """Compute the cut leaves from a given HalfSpace object.

    For the half plane, we find all the cut leaves in the slice of size bin_width,
    and compute the quality of the cut (see docstring of find_cut_leaves for details).
    If the quality is positive, a cut is considered valid, and the cut leaves are returned.

    Args:
        half_space (planes.HalfSpace): half space to search cut points
        morphology (morphio.Morphology): morphology
        bin_width: the bin width
        percentile_threshold: the minimum percentile of leaves counts in bins

    Returns:
        leaves: ndarray of dim (n, 3) with cut leaves coordinates
        quality: quality for these cut leaves
    """
    # get the cut leaves
    leaves = np.array([section for section in morphology.iter() if not section.children])
    leaves_coord = np.array([leaf.points[-1, COLS.XYZ] for leaf in leaves])
    cut_filter = half_space.distance(leaves_coord) < bin_width
    cut_leaves = leaves[cut_filter]

    # compute the min cut leave given the percentile
    projected_uncut_leaves = half_space.project_on_directed_normal(leaves_coord[~cut_filter])
    if projected_uncut_leaves.size == 0:
        return None, None

    _min, _max = min(projected_uncut_leaves), max(projected_uncut_leaves)
    bins = np.arange(_min, _max, bin_width)
    _dig = np.digitize(projected_uncut_leaves, bins)
    leaves_threshold = np.percentile(np.unique(_dig, return_counts=True)[1], percentile_threshold)

    quality = len(cut_leaves) - leaves_threshold
    if quality > 0:
        return leaves_coord[cut_filter], quality
    else:
        return None, None


def find_cut_leaves(
    morph,
    bin_width=3,
    percentile_threshold=70.0,
    searched_axes=("Z",),
    searched_half_spaces=(-1, 1),
):
    """Find all cut leaves for cuts with strong signal for real cut.

    The algorithm works as follow. Given the searched_axes and searched_half_spaces,
    a list of candidate cuts is created, consisting of a slice with bin_width adjusted to the most
    extreme points of the morphology in the direction of serched_axes/serached_half_spaces.
    Each cut contains a set of leaves, which are considered as cut leaves if their quality
    is positive. The quality of a cut is defined the number of leaves in the cut minus the
    'percentile_threshold' percentile of the distribution of the number of leaves in all other
    slices of bin_width size of the morphology. More explicitely, if a cut has more leaves than most
    of other possible cuts of the same size, it is likely to be a real cut from an invitro slice.

    Note that all cuts can be valid, thus cut leaves can be on both sides.

    Args:
        morph (morphio.Morphology): morphology
        bin_width: the bin width
        percentile_threshold: the minimum percentile of leaves counts in bins
        searched_axes: x, y or z. Specify the half space for which to search the cut leaves
        searched_half_spaces: A negative value means the morphology lives
                on the negative side of the plane, and a positive one the opposite.
    Returns:
        ndarray: cut leaves
        list: list of qualities in dicts with axis and side for each
    """
    # create half spaces
    searched_axes = [axis.upper() for axis in searched_axes]
    half_spaces = [
        HalfSpace(int(axis == "X"), int(axis == "Y"), int(axis == "Z"), 0, upward=(side > 0))
        for axis, side in product(searched_axes, searched_half_spaces)
    ]

    # set the half space coef_d as furthest morphology point
    for half_space, (axis, side) in zip(half_spaces, product(searched_axes, searched_half_spaces)):
        half_space.coefs[3] = -side * np.min(
            half_space.project_on_directed_normal(morph.points), axis=0
        )

    # find the cut leaves
    cuts = [
        _get_cut_leaves(half_space, morph, bin_width, percentile_threshold)
        for half_space in half_spaces
    ]

    # return only cut leaves of half spaces with valid cut
    _leaves = [leave for leave, _ in cuts if leave is not None]
    leaves = np.vstack(_leaves) if _leaves else np.array([])
    qualities = [
        {"axis": axis, "side": side, "quality": np.around(quality, 3)}
        for (_, quality), (axis, side) in zip(cuts, product(searched_axes, searched_half_spaces))
        if quality is not None
    ]
    return leaves, qualities



from pathlib import Path
from tempfile import TemporaryDirectory
import numpy as np

from morphio import Morphology
from nose.tools import assert_raises
from numpy.testing import assert_array_equal, assert_equal

from morph_tool.utils import iter_morphology_files
from neuror.sanitize import CorruptedMorphology, fix_non_zero_segments, sanitize, sanitize_all
from neuror.sanitize import annotate_neurolucida, annotate_neurolucida_all

PATH = Path(__file__).parent / 'data'


def test_fix_non_zero_segments():
    neuron = fix_non_zero_segments(Path(PATH, 'simple-with-duplicates.asc'))
    assert_equal(len(neuron.root_sections), 1)
    assert_array_equal(neuron.section(0).points,
                       [[0., 0., 0.],
                        [1., 1., 0.],
                        [2., 0., 0.],
                        [3., 0., 0.]])


def test_sanitize():
    with TemporaryDirectory('test-sanitize') as tmp_folder:
        output_path = Path(tmp_folder, 'sanitized.asc')
        sanitize(Path(PATH, 'simple-with-duplicates.asc'), output_path)
        neuron = Morphology(output_path)
        assert_equal(len(neuron.root_sections), 1)
        assert_array_equal(neuron.section(0).points,
                           [[0., 0., 0.],
                            [1., 1., 0.],
                            [2., 0., 0.],
                            [3., 0., 0.]])

        for input_morph, expected_exception in [
                ('no-soma.asc',
                 '{} has an invalid or no soma'.format(Path(PATH, 'no-soma.asc'))),

                ('neurite-with-multiple-types.swc',
                 ('{} has a neurite whose type changes along the way\n'
                  'Child section (id: 5) has a different type (SectionType.basal_dendrite) '
                  'than its parent (id: 3) (type: SectionType.axon)').format(
                      Path(PATH, 'neurite-with-multiple-types.swc')))
        ]:
            with assert_raises(CorruptedMorphology) as cm:
                sanitize(PATH / input_morph, Path(tmp_folder, 'output.asc'))
            assert_equal(str(cm.exception), expected_exception)

        out_path = Path(tmp_folder, 'output.asc')
        sanitize(PATH / 'negative-diameters.asc', out_path)
        assert_equal(next(Morphology(out_path).iter()).diameters, [2, 2, 0, 2])


def test_sanitize_all():
    with TemporaryDirectory('test-sanitize') as tmp_folder:
        tmp_folder = Path(tmp_folder)
        sanitize_all(PATH / 'input-sanitize-all', tmp_folder)

        assert_array_equal(list(sorted(tmp_folder.rglob('*.asc'))),
                           [tmp_folder / 'a.asc',
                            tmp_folder / 'sub-folder/sub-sub-folder/c.asc'])

    with TemporaryDirectory('test-sanitize') as tmp_folder:
        tmp_folder = Path(tmp_folder)
        sanitize_all(PATH / 'input-sanitize-all', tmp_folder, nprocesses=2)

        assert_array_equal(list(sorted(tmp_folder.rglob('*.asc'))),
                           [tmp_folder / 'a.asc',
                            tmp_folder / 'sub-folder/sub-sub-folder/c.asc'])


def test_error_annotation():
    annotation, summary, markers = annotate_neurolucida(
        Path(PATH, 'test-error-detection/error-morph.asc')
    )
    assert_equal(summary, {'fat end': 1,
                           'zjump': 1,
                           'narrow start': 1,
                           'dangling': 1,
                           'Multifurcation': 1})
    assert_equal(markers, [{'name': 'fat end', 'label': 'Circle3', 'color': 'Blue',
                            'data': [(7, np.array([[-5., -4.,  0., 20.]], dtype=np.float32))]},
                           {'name': 'zjump', 'label': 'Circle2', 'color': 'Green',
                            'data': [(2, [np.array([0., 5., 0., 1.], dtype=np.float32),
                                          np.array([0.,  5., 40.,  1.], dtype=np.float32)])]},
                           {'name': 'narrow start', 'label': 'Circle1', 'color': 'Blue',
                            'data': [(0, np.array([[0., 5., 0., 1.]], dtype=np.float32))]},
                           {'name': 'dangling', 'label': 'Circle6', 'color': 'Magenta',
                            'data': [(5, [np.array([10., -20.,  -4., 1.], dtype=np.float32)])]},
                           {'name': 'Multifurcation', 'label': 'Circle8', 'color': 'Yellow',
                            'data': [(0, np.array([[0., 5., 0., 1.]], dtype=np.float32))]}])


def test_error_annotation_all():

    input_dir = Path(PATH, 'test-error-detection')
    # this ensure morphs are ordered as expected
    morph_paths = sorted([str(morph) for morph in iter_morphology_files(input_dir)])
    annotations, summaries, markers = annotate_neurolucida_all(morph_paths)
    assert_equal(summaries, {str(morph_paths[0]): {'fat end': 1,
                                                   'zjump': 1,
                                                   'narrow start': 1,
                                                   'dangling': 1,
                                                   'Multifurcation': 1},
                             str(morph_paths[1]): {}})
    assert_equal(markers, {
        str(morph_paths[0]): [{'name': 'fat end', 'label': 'Circle3', 'color': 'Blue',
                               'data': [(7, np.array([[-5., -4.,  0., 20.]], dtype=np.float32))]},
                              {'name': 'zjump', 'label': 'Circle2', 'color': 'Green',
                               'data': [(2, [np.array([0., 5., 0., 1.], dtype=np.float32),
                                             np.array([0.,  5., 40.,  1.], dtype=np.float32)])]},
                              {'name': 'narrow start', 'label': 'Circle1', 'color': 'Blue',
                               'data': [(0, np.array([[0., 5., 0., 1.]], dtype=np.float32))]},
                              {'name': 'dangling', 'label': 'Circle6', 'color': 'Magenta',
                               'data': [(5, [np.array([10., -20.,  -4., 1.], dtype=np.float32)])]},
                              {'name': 'Multifurcation', 'label': 'Circle8', 'color': 'Yellow',
                               'data': [(0, np.array([[0., 5., 0., 1.]], dtype=np.float32))]}],
        str(morph_paths[1]): []})



from pathlib import Path

from morph_tool import diff
from morphio import SectionType
from neurom import load_neuron
from nose.tools import assert_raises, ok_
from numpy.testing import assert_array_equal, assert_equal

import neuror.axon as test_module

DATA_PATH = Path(__file__).parent / 'data'

SIMPLE_PATH = DATA_PATH / 'simple.swc'
SLICE_PATH = DATA_PATH / 'neuron-slice.h5'
SIMPLE = load_neuron(SIMPLE_PATH)
SLICE = load_neuron(SLICE_PATH)


def test_tree_distance():
    child = SLICE.neurites[0].root_node.children[0]
    assert_equal(test_module._tree_distance(child, SLICE.neurites[0].root_node), 1)

    child1, child2 = SLICE.neurites[0].root_node.children
    assert_equal(test_module._tree_distance(child1, child2), 1)
    assert_equal(test_module._tree_distance(child2, child1), 1)

    assert_equal(test_module._tree_distance(child1.children[0], child2.children[0]), 3)


    assert_raises(Exception, test_module._tree_distance, SLICE.neurites[0].root_node,
                  SLICE.neurites[1].root_node)

def test__downstream_pathlength():
    root = SIMPLE.root_sections[0]
    assert_equal(test_module._downstream_pathlength(root), 16)
    assert_equal(test_module._downstream_pathlength(root.children[0]), 5)


def test__similar_section():
    root = SIMPLE.root_sections[0]
    similar = test_module._similar_section(SIMPLE.root_sections, root)
    assert_equal(similar.id, root.id)


def test__sort_intact_sections_by_score():
    root = SIMPLE.root_sections[0]

    res = test_module._sort_intact_sections_by_score(root, root, SIMPLE.root_sections)
    assert_equal(res[0].id, root.id)


def test__repair():
    neuron = load_neuron(Path(DATA_PATH, 'valid.h5'))
    axon = neuron.root_sections[0]
    assert_equal(axon.type, SectionType.axon)
    test_module.repair(neuron, axon, [axon], [axon], set(), y_extent=10000)
    assert_equal(len(axon.children), 1)
    assert_array_equal(axon.children[0].points[0], axon.points[-1])
    ok_(not diff(neuron, DATA_PATH / 'axon-repair.h5'))


def test__repair_no_intact_axon():
    filename = Path(DATA_PATH, 'valid.h5')
    neuron = load_neuron(filename)
    axon = neuron.root_sections[0]
    used_axon_branches = set()
    test_module.repair(neuron, axon, [], [axon], used_axon_branches, y_extent=10000)
    # There should not be any repair
    ok_(not diff(neuron, filename))



'''The axon repair module'''
import logging

import morphio
import numpy as np
from morph_tool.transform import align, translate
from neurom import COLS
from neurom.features.sectionfunc import branch_order, strahler_order

from neuror.utils import section_length

L = logging.getLogger('neuror')


def _tree_distance(sec1, sec2):
    '''
    Returns the number of sections between the 2 sections

    Reimplementation of:
    https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_axon.cpp#n35

    raises: if both sections are not part of the same neurite

    Note:
    I think the implementation of tree distance is true to the original
    but I would expect the tree distance of 2 children with the same parent to be 2 and not 1
    Because in the current case, (root, child1) and (child1, child2) have the
    same tree distance and it should probably not be the case
    '''
    original_sections = (sec1, sec2)
    dist = 0
    while True:
        diff = branch_order(sec1) - branch_order(sec2)
        if diff == 0:
            break

        if diff > 0:
            sec1 = sec1.parent
            dist += 1
        else:
            sec2 = sec2.parent
            dist += 1

    if sec1.id == sec2.id:
        return dist

    dist -= 1
    while sec1.id != sec2.id:
        sec1 = sec1.parent
        sec2 = sec2.parent
        dist += 2
        if None in {sec1, sec2}:
            raise Exception('Sections {} and {} are not part of the same neurite'.format(
                original_sections[0], original_sections[1]))

    return dist


def _downstream_pathlength(section):
    '''The sum of this section and its descendents's pathlengths

    Reimplementation of the C++ function "children_length":

    https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/morphstats.cpp#n112
    '''
    ret = section_length(section)
    for child in section.children:
        ret += _downstream_pathlength(child)
    return ret


def _similar_section(intact_axons, section):
    '''Use the "mirror" technique of BlueRepairSDK to find out the similar section

    https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_axon.cpp#n83

    Note:
    I have *absolutely* no clue why sorting by this metric
    '''
    dists = list()
    for root in intact_axons:
        origin = root.points[0]
        origin_cut = section.points[0]
        diff = origin_cut - origin
        diff[COLS.X] = origin_cut[COLS.X] + origin[COLS.X]
        dists.append(np.linalg.norm(diff))

    return intact_axons[np.argmin(dists)]


def _sort_intact_sections_by_score(section, similar_section, axon_branches):
    '''Returns an array of sections sorted by their score'''
    reference = _downstream_pathlength(similar_section) - section_length(section)

    def score(branch):
        '''The score. The interpretation is something like the absolute difference in
        remaining children length'''
        return -abs(reference - _downstream_pathlength(branch))
    return sorted(axon_branches, key=score)


def repair(morphology, section, intact_sections, axon_branches, used_axon_branches, y_extent):
    '''Axonal repair

    Reimplementation of:
    https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/repair.cpp#n727

    1) Find the most similar section in INTACT_SECTIONS list to SECTION
    2) Sort AXON_BRANCHES according to a similarity score to the section found at step 1
    3) Loop through the sorted AXON_BRANCHES to find a section with same strahler orders
       and that, when appended, does not extend further than Y_EXTENT
    4) Append the first section that meets the conditions of step 3)
    5) Mark this section as used and do not re-use it

    Args:
        morphology (neurom.Neuron): the morphology to repair
        section (neurom.core._neuron.Section): the section to repair
        intact_sections (List[Section]): a list of all sections from this morphology
            that are part of an intact subtree. Note: these section won't be grafted
        axon_branches (List[Section]): a list a intact sections coming from donor morphologies
            These are the sections that will be appended

    ..note::
       The original code used to have more parameters. In the context of the bbp-morphology-workflow
       it seems that some of the parameters were always used with the same value. This
       reimplementation assumes the following BlueRepairSDK options:

        - --overlap=true
        - --incremental=false
        - --restrict=true
        - --distmethod=mirror'
    '''

    if not intact_sections:
        L.debug("No intact axon found. Not repairing!")
        return

    similar = _similar_section(intact_sections, section)
    branch_pool = _sort_intact_sections_by_score(section, similar, axon_branches)

    strahler_orders = {intact_section: strahler_order(intact_section)
                       for intact_section in intact_sections + branch_pool}

    L.debug('Branch pool count: %s', len(branch_pool))
    for branch in branch_pool:
        if (branch in used_axon_branches or strahler_orders[similar] != strahler_orders[branch]):
            continue

        L.debug("Pasting axon branch with ID %s", branch.id)

        end_point = section.points[-1, COLS.XYZ]
        appended = section.append_section(branch)
        translation = section.points[-1, COLS.XYZ] - appended.points[0, COLS.XYZ]
        align(appended, translation)
        translate(appended, translation)

        # Make sure the child section first point is exactly the end point of the parent section
        appended_points = np.copy(appended.points)
        appended_points[0] = end_point
        appended.points = appended_points

        if any(np.any(section.points[:, COLS.Y] > y_extent) for section in appended.iter()):
            L.debug("Discarded, exceeds y-limit")
            morphology.delete_section(appended)
        else:
            L.debug('Section appended')
            used_axon_branches.add(branch)
            return

    morphio.set_ignored_warning(morphio.Warning.wrong_duplicate, False)



import json
from pathlib import Path
from tempfile import TemporaryDirectory

from click.testing import CliRunner
from numpy.testing import assert_array_almost_equal

from neuror.cut_plane.cli import cli

DATA = Path(__file__).parent.parent / 'data'

NEURON_SLICE = str(DATA / 'Neuron_slice.h5')


def test_cli_compute_file():
    runner = CliRunner()

    # vanilla test
    result = runner.invoke(cli, ['compute', 'file', NEURON_SLICE])
    if result.exit_code:
        raise AssertionError(result.exception)

    # test writing output
    with TemporaryDirectory(prefix='test_cli_compute_file') as temp_dir:
        json_file = str(Path(temp_dir, 'plane.json'))
        result = runner.invoke(cli, ['compute', 'file', NEURON_SLICE,
                                     '-o', json_file])
        if result.exit_code:
            raise AssertionError(result.exception)

        with open(json_file) as f:
            data = json.load(f)
            assert_array_almost_equal(data['cut-plane']['d'], -48.68020515427703)

    # test forcing plane
    with TemporaryDirectory(prefix='test_cli_compute_file') as temp_dir:
        json_file = str(Path(temp_dir, 'plane.json'))
        result = runner.invoke(cli, ['compute', 'file', NEURON_SLICE,
                                     '-o', json_file,
                                     '--plane', 'y'])
        if result.exit_code:
            raise AssertionError(result.exception)

        with open(json_file) as f:
            data = json.load(f)
            assert_array_almost_equal([data['cut-plane']['a'],
                                       data['cut-plane']['b'],
                                       data['cut-plane']['c']],
                                      [0, 1, 0])

    # test forcing plane and position
    with TemporaryDirectory(prefix='test_cli_compute_file') as temp_dir:
        json_file = str(Path(temp_dir, 'plane.json'))
        result = runner.invoke(cli, ['compute', 'file', NEURON_SLICE,
                                     '-o', json_file,
                                     '--plane', 'y',
                                     '--position', 40])
        if result.exit_code:
            raise AssertionError(result.exception)

        with open(json_file) as f:
            data = json.load(f)
            assert_array_almost_equal([data['cut-plane']['a'],
                                       data['cut-plane']['b'],
                                       data['cut-plane']['c'],
                                       data['cut-plane']['d']],
                                      [0, 1, 0, -40])


def test_cli_compute_folder():
    runner = CliRunner()

    with TemporaryDirectory(prefix='cut-plane-compute-folder') as temp_dir:
        result = runner.invoke(cli, ['compute', 'folder', str(DATA), temp_dir])
    if result.exit_code:
        raise AssertionError(result.exception)



import json
from pathlib import Path
from tempfile import TemporaryDirectory

from click.testing import CliRunner
from nose.tools import assert_dict_equal, assert_equal

from neuror.cut_plane.cli import cli

DATA = Path(__file__).parent.parent / 'data'


def test_join():
    with TemporaryDirectory(prefix='cut-plane-join') as temp_dir:
        out_filename = str(Path(temp_dir, 'plane.json'))

        runner = CliRunner()
        args = ['join',
                out_filename,
                str(DATA / 'plane1.json'),
                str(DATA / 'plane2.json'),
                str(DATA / 'plane3.json')]

        result = runner.invoke(cli, args)
        assert result.exit_code == 0

        with open(out_filename) as f:
            actual = json.load(f)

        with open(out_filename) as f:
            expected = json.load(f)

        assert_equal(len(actual), len(expected))
        for act, exp in zip(actual, expected):
            assert_dict_equal(act, exp)



'''This module defines classes PlaneEquation and CutPlane'''
import numpy as np
from neurom import COLS
from pyquaternion import Quaternion

A, B, C, D = range(4)
ABC = slice(0, 3)


def _get_displaced_pos(pos, quat, size_multiplier, axis):
    """ Compute the shifted position wrt the quaternion and axis """
    return pos + size_multiplier * np.array(quat.rotate(axis))


class PlaneEquation(object):
    '''This class defines the equation of a plane.
    It is a mathematical object which is domain agnostic.

    a X + b Y + c Z + d = 0
    '''

    def __init__(self, a, b, c, d):
        if (a, b, c) == (0, 0, 0):
            raise ValueError('Cannot define a plane with a == b == c == 0')
        self.coefs = np.array([a, b, c, d], dtype=np.float)

    @classmethod
    def from_dict(cls, data):
        '''Instantiate object from dict like:
             {"a": 1, "b": 2, "c": 0, "d": 10}
        '''
        return cls(data['a'], data['b'], data['c'], data['d'])

    @classmethod
    def from_rotations_translations(cls, transformations):
        '''Factory method to build the plane equation from the rotation and translation
        provided by the cut_plane.viewer

        Args:
            transformations: An array [rot_x, rot_y, rot_z, transl_x, transl_y, transl_z]
        '''
        rot_x, rot_y, rot_z, transl_x, transl_y, transl_z = transformations
        qx = Quaternion(axis=[1, 0, 0], angle=rot_x / 180. * np.pi).unit
        qy = Quaternion(axis=[0, 1, 0], angle=rot_y / 180. * np.pi).unit
        qz = Quaternion(axis=[0, 0, 1], angle=rot_z / 180. * np.pi).unit
        quat = qx * qy * qz
        normal_vector = _get_displaced_pos([0, 0, 0], quat, 100, (0, 0, 1))
        transl = np.array([transl_x, transl_y, transl_z])
        return cls(normal_vector[0],
                   normal_vector[1],
                   normal_vector[2],
                   -transl.dot(normal_vector))

    def distance(self, points):
        '''Returns an array containing the distance between the plane and each point'''
        points = np.array(points)
        u = self.coefs[ABC].copy()
        return np.abs(points.dot(u) + self.coefs[D]) / np.linalg.norm(u)

    def __str__(self):
        return '({0}) * X + ({1}) * Y + ({2}) * Z + ({3}) = 0'.format(*self.coefs)

    def to_json(self):
        '''Returns a dict for json serialization'''
        return dict(a=self.coefs[A], b=self.coefs[B], c=self.coefs[C], d=self.coefs[D],
                    comment="Equation: a*X + b*Y + c*Z + d = 0")

    @property
    def normal(self):
        '''Returns the vector orthogonal to the plane'''
        return self.coefs[ABC] / np.linalg.norm(self.coefs[ABC])

    def project_on_normal(self, points):
        '''Returns the points projected on the normal vector'''
        if self.coefs[A]:
            point_in_the_plane = np.array([-self.coefs[D] / self.coefs[A], 0, 0])
        elif self.coefs[B]:
            point_in_the_plane = np.array([0, -self.coefs[D] / self.coefs[B], 0])
        else:  # __init__ would have raised if c was 0 as well
            point_in_the_plane = np.array([0, 0, -self.coefs[D] / self.coefs[C]])

        return points[:, COLS.XYZ].dot(self.normal) - point_in_the_plane.dot(self.normal)

    def count_near_plane(self, points, bin_width):
        '''Return the number of points in ]-bin_width, 0] and ]0, bin_width]'''
        points = self.project_on_normal(points)
        n_left = len(points[(points > -bin_width) & (points <= 0)])
        n_right = len(points[(points > 0) & (points <= bin_width)])
        return n_left, n_right


class HalfSpace(PlaneEquation):
    '''
    A mathematical half-space: https://en.wikipedia.org/wiki/Half-space_(geometry)

    a X + b Y + c Z + d > 0 if upward == True
    a X + b Y + c Z + d < 0 else
    '''

    def __init__(self, a, b, c, d, upward):
        super().__init__(a, b, c, d)
        self.upward = upward

    def to_json(self):
        '''Returns a dict for json serialization'''
        return dict(a=self.coefs[A], b=self.coefs[B], c=self.coefs[C], d=self.coefs[D],
                    upward=self.upward,
                    comment="Equation: a*X + b*Y + c*Z + d {} 0".format(
                        '>' if self.upward else '<'))

    def project_on_directed_normal(self, points):
        '''Project on the normal oriented toward the inside of the half-space'''
        points = self.project_on_normal(points)
        if self.upward:
            return points
        else:
            return -points



'''App to find cut planes with arbitrary cut planes orientations
with the help of a manual hint

Related to https://bbpteam.epfl.ch/project/issues/browse/NGV-85
'''
import json

import neurom as nm
import numpy as np
from neurom import load_neuron
from neurom.geom import bounding_box
from pyquaternion import Quaternion

from neuror.cut_plane.detection import CutPlane, _minimize
from neuror.cut_plane.planes import _get_displaced_pos

try:
    import dash
    import dash_core_components as dcc
    import dash_html_components as html
    from dash.dependencies import Input, Output, State
    from plotly_helper.neuron_viewer import NeuronBuilder
except ImportError as e:
    raise ImportError(
        'neuror[plotly] is not installed.'
        ' Please install it by doing: pip install neuror[plotly]') from e


# Copy of: https://codepen.io/chriddyp/pen/bWLwgP.css
external_stylesheets = ['dash.css']
app = dash.Dash(__name__, external_stylesheets=external_stylesheets)

BIN_WIDTH = 10


class NumpyEncoder(json.JSONEncoder):
    '''JSON encoder that handles numpy types

    In python3, numpy.dtypes don't serialize to correctly, so a custom converter
    is needed.
    '''

    def default(self, o):  # pylint: disable=method-hidden
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.ndarray):
            return o.tolist()
        return json.JSONEncoder.default(self, o)


def create_plane(pos, quat):
    """ Create a 3d plane using a center position and a quaternion for orientation

    Args :
        pos: x,y,z position of the plane's center (array([x,y,z]))
        quat: quaternion representing the orientations (Quaternion)
        size_multiplier: plane size in space coordinates (float)
        opacity: set the opacity value (float)

    Returns :
        A square surface to the plotly format
    """

    length = np.linalg.norm(BBOX[1] - BBOX[0]) / 2.
    positif_x = _get_displaced_pos(pos, quat, length, (1, 0, 0))
    positif_y = _get_displaced_pos(pos, quat, length, (0, 1, 0))
    negatif_x = _get_displaced_pos(pos, quat, length, (-1, 0, 0))
    negatif_y = _get_displaced_pos(pos, quat, length, (0, -1, 0))

    x = [[positif_x[0], positif_y[0]], [negatif_y[0], negatif_x[0]]]
    y = [[positif_x[1], positif_y[1]], [negatif_y[1], negatif_x[1]]]
    z = [[positif_x[2], positif_y[2]], [negatif_y[2], negatif_x[2]]]

    return dict(
        z=z,
        x=x,
        y=y,
        showscale=False,
        type='surface',
        surfacecolor=['green', 'green'], opacity=1
    )


ROT_X, ROT_Y, ROT_Z = 4, 45, -21
TRANSL_X, TRANSL_Y, TRANSL_Z = 0, 0, 61

app.layout = html.Div(children=[
    html.Pre(id='click-data'),
    html.Pre(id='neuron'),

    html.Pre(id='optimized'),

    html.Div(
        [
            "Rotate the plane until it is aligned with the cut plane, then click optimize, "
            "then click export. "
            'On the histogram, all points should fall on only one side of the red line',
            html.Div(
                [
                    dcc.Graph(id='graph'),
                ],
                style={'width': '70%', 'display': 'inline-block'},
            ),

            html.Div(
                [
                    dcc.Graph(id='bar'),
                    html.Div([
                        html.Button('Optimize', id='button'),
                        dcc.Checklist(id='hide-plane',
                                      options=[
                                          {'label': 'Hide plane', 'value': 'hidden'}
                                      ],
                                      value=[]),

                        "Rotations are with respect to the frame attached to the plane.",
                        html.Div(
                            [html.Div(['Rotate X:'], id='output-x-rotate'),
                             dcc.Input(id='rotate-x-slider', type='number', value=ROT_X), ]
                        ),

                        html.Div(
                            [html.Div(['Rotate Y:'], id='output-y-rotate'),
                             dcc.Input(id='rotate-y-slider', type='number', value=ROT_Y), ],
                        ),

                        html.Div(
                            [html.Div(['Rotate Z:'], id='output-z-rotate'),
                             dcc.Input(id='rotate-z-slider', type='number', value=ROT_Z), ]
                        ),

                        html.Div(
                            [html.Div(['Translate X:'], id='output-x-translate'),
                             dcc.Input(id='translate-x-slider', type='number', value=TRANSL_X)]
                        ),

                        html.Div(
                            [html.Div(['Translate Y:'], id='output-y-translate'),
                             dcc.Input(id='translate-y-slider', type='number', value=TRANSL_Y), ],
                        ),

                        html.Div(
                            [html.Div(['Translate Z:'], id='output-z-translate'),
                             dcc.Input(id='translate-z-slider',
                                       type='number', value=TRANSL_Z, min=-10000), ]
                        ),

                        dcc.Input(id='export-path-input', type='text', value='/tmp/cut-plane.json'),
                        html.Button('export', id='export')], style={'margin-left': '100px'})
                ],
                style={'width': '30%', 'display': 'inline-block', 'vertical-align': 'top'}
            ),
        ],
        style={'width': '100%', 'display': 'inline'})
]
)

NEURON = FIGURE = BBOX = None


def set_neuron(filename):
    '''Globally loads the neuron'''
    global NEURON, FIGURE, BBOX  # pylint: disable=global-statement
    NEURON = load_neuron(filename)
    FIGURE = NeuronBuilder(NEURON, '3d').get_figure()
    BBOX = bounding_box(NEURON)


@app.callback(
    Output('graph', 'figure'),
    [
        Input('rotate-x-slider', 'value'),
        Input('rotate-y-slider', 'value'),
        Input('rotate-z-slider', 'value'),
        Input('translate-x-slider', 'value'),
        Input('translate-y-slider', 'value'),
        Input('translate-z-slider', 'value'),
        Input('hide-plane', 'value'),
    ],
    [State('graph', 'relayoutData')]

)
def display_click_data(rot_x, rot_y, rot_z, transl_x, transl_y, transl_z, hide, layout):
    '''callback that redraw everything when sliders are changed'''
    qx = Quaternion(axis=[1, 0, 0], angle=rot_x / 180. * np.pi)
    qy = Quaternion(axis=[0, 1, 0], angle=rot_y / 180. * np.pi)
    qz = Quaternion(axis=[0, 0, 1], angle=rot_z / 180. * np.pi)
    plane = create_plane([transl_x, transl_y, transl_z], qx * qy * qz)

    FIGURE['data'] = [x for x in FIGURE['data'] if not isinstance(x, dict)]
    if not hide:
        FIGURE['data'].append(plane)
        FIGURE['layout']['autosize'] = False
        FIGURE['layout']['height'] = 1500

    if layout and layout.get('scene.camera'):
        FIGURE['layout']['scene']['camera'] = layout['scene.camera']

    return FIGURE


@app.callback(
    dash.dependencies.Output('bar', 'figure'),
    [
        Input('rotate-x-slider', 'value'),
        Input('rotate-y-slider', 'value'),
        Input('rotate-z-slider', 'value'),
        Input('translate-x-slider', 'value'),
        Input('translate-y-slider', 'value'),
        Input('translate-z-slider', 'value'),
    ])
def update_output(rot_x, rot_y, rot_z, transl_x, transl_y, transl_z):
    '''Update histo when sliders are changed'''
    transformations = [rot_x, rot_y, rot_z, transl_x, transl_y, transl_z]
    bin_width = 10
    cut_plane = CutPlane.from_rotations_translations(transformations, NEURON, bin_width)
    hist, binning = cut_plane.histogram()

    binning += bin_width / 2.

    return {
        'data': [
            {'x': binning, 'y': hist, 'type': 'bar', 'name': 'SF'},
        ],
        'layout': {
            'title': 'Distance to plane distribution',
            'xaxis': {
                'title': 'Distance to plane (um)'
            },
            'yaxis': {
                'title': 'Counts'
            },
            'shapes': [
                # Line Vertical
                {
                    'type': 'line',
                    'x0': 0,
                    'y0': 0,
                    'x1': 0,
                    'y1': np.max(hist),
                    'line': {
                        'color': 'red',
                        'width': 3,
                    },
                }
            ]
        }
    }


@app.callback(
    dash.dependencies.Output('optimized', 'data-*'),
    [Input('button', 'n_clicks')],
    [
        State('rotate-x-slider', 'value'),
        State('rotate-y-slider', 'value'),
        State('rotate-z-slider', 'value'),
        State('translate-x-slider', 'value'),
        State('translate-y-slider', 'value'),
        State('translate-z-slider', 'value'),
    ]
)
def optimize(n_clicks, rot_x, rot_y, rot_z, transl_x, transl_y, transl_z):
    '''Optimize cut plane parameters'''
    if not n_clicks:
        return rot_x, rot_y, rot_z, transl_x, transl_y, transl_z
    points = np.array([point
                       for neurite in (NEURON.neurites or [])
                       for section in nm.iter_sections(neurite)
                       for point in section.points])
    params = rot_x, rot_y, rot_z, transl_x, transl_y, transl_z
    result = _minimize(params, points, bin_width=BIN_WIDTH)
    return result


@app.callback(
    Output('rotate-x-slider', 'value'),
    [Input('optimized', 'data-*')])
def update_post_optim_x_rotate(params):
    '''callback'''
    return params[0]


@app.callback(
    Output('rotate-y-slider', 'value'),
    [Input('optimized', 'data-*')])
def update_post_optim_y_rotate(params):
    '''callback'''
    return params[1]


@app.callback(
    Output('rotate-z-slider', 'value'),
    [Input('optimized', 'data-*')])
def update_post_optim_z_rotate(params):
    '''callback'''
    return params[2]


@app.callback(
    Output('translate-x-slider', 'value'),
    [Input('optimized', 'data-*')])
def update_post_optim_x_translate(params):
    '''callback'''
    return params[3]


@app.callback(
    Output('translate-y-slider', 'value'),
    [Input('optimized', 'data-*')])
def update_post_optim_y_translate(params):
    '''callback'''
    return params[4]


@app.callback(
    Output('translate-z-slider', 'value'),
    [Input('optimized', 'data-*')])
def update_post_optim_z_translate(params):
    '''callback'''
    return params[5]


@app.callback(
    dash.dependencies.Output('click-data', 'children'),
    [Input('export', 'n_clicks')],
    [State('rotate-x-slider', 'value'),
     State('rotate-y-slider', 'value'),
     State('rotate-z-slider', 'value'),
     State('translate-x-slider', 'value'),
     State('translate-y-slider', 'value'),
     State('translate-z-slider', 'value'),
     State('export-path-input', 'value')])
def export(n_clicks, rot_x, rot_y, rot_z, transl_x, transl_y, transl_z, output_path):
    '''Write the final file cut-plane.json to disk'''
    if not n_clicks:
        return
    plane = CutPlane.from_rotations_translations(
        [rot_x, rot_y, rot_z, transl_x, transl_y, transl_z],
        NEURON, BIN_WIDTH)
    payload = [plane.to_json()]
    with open(output_path, 'w') as f:
        json.dump(payload, f, cls=NumpyEncoder)



'''Module for the sanitization of raw morphologies.'''
import logging
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from tqdm import tqdm

import numpy as np
from morphio import MorphioError, SomaType, set_maximum_warnings
from morphio.mut import Morphology  # pylint: disable=import-error
from neurom.check import neuron_checks as nc
from neurom.check import CheckResult
from neurom.apps.annotate import annotate
from neurom import load_neuron

L = logging.getLogger('neuror')
_ZERO_LENGTH = 1e-4


class CorruptedMorphology(Exception):
    '''Exception for morphologies that should not be used'''


def iter_morphologies(folder):
    '''Recursively yield morphology files in folder and its sub-directories.'''
    return (path for path in folder.rglob('*') if path.suffix.lower() in {'.swc', '.h5', '.asc'})


def sanitize(input_neuron, output_path):
    '''Sanitize one morphology.

    - ensures it can be loaded with MorphIO
    - raises if the morphology has no soma or of invalid format
    - removes unifurcations
    - set negative diameters to zero
    - raises if the morphology has a neurite whose type changes along the way
    - removes segments with near zero lengths (shorter than 1e-4)

    Args:
        input_neuron (str|pathlib.Path|morphio.Morphology|morphio.mut.Morphology): input neuron
        output_path (str|pathlib.Path): output name
    '''
    neuron = Morphology(input_neuron)

    if neuron.soma.type == SomaType.SOMA_UNDEFINED:  # pylint: disable=no-member
        raise CorruptedMorphology('{} has an invalid or no soma'.format(input_neuron))

    neuron.remove_unifurcations()

    for section in neuron.iter():
        section.diameters = np.clip(section.diameters, 0, None)

    for root in neuron.root_sections:  # pylint: disable=not-an-iterable
        for section in root.iter():
            if section.type != root.type:
                raise CorruptedMorphology(f'{input_neuron} has a neurite whose type changes along '
                                          'the way\n'
                                          f'Child section (id: {section.id}) has a different type '
                                          f'({section.type}) than its parent (id: '
                                          f'{section.parent.id}) (type: {section.parent.type})')

    fix_non_zero_segments(neuron).write(str(output_path))


def _sanitize_one(path, input_folder, output_folder):
    '''Function to be called by sanitize_all to catch all exceptions
    and return path if in error

    Since Pool.imap_unordered only supports one argument, the argument
    is a tuple: (path, input_folder, output_folder).
    '''
    relative_path = path.relative_to(input_folder)
    output_dir = output_folder / relative_path.parent
    if not output_dir.exists():
        # exist_ok=True since there can be race conditions because of Pool
        output_dir.mkdir(parents=True, exist_ok=True)
    try:
        sanitize(path, output_dir / path.name)
    except (MorphioError, CorruptedMorphology):
        return str(path)
    else:
        return None


def sanitize_all(input_folder, output_folder, nprocesses=1):
    '''Sanitize all morphologies in input_folder and its sub-directories.

    See :func:`~neuror.sanitize.sanitize` for more information on the sanitization process.

    Args:
        input_folder (str|pathlib.Path): input neuron
        output_folder (str|pathlib.Path): output name

    .. note:: the sub-directory structure is maintained.
    '''
    set_maximum_warnings(0)

    morphologies = list(iter_morphologies(Path(input_folder)))
    func = partial(_sanitize_one, input_folder=input_folder, output_folder=output_folder)
    if nprocesses == 1:
        results = map(func, morphologies)
    else:
        results = Pool(nprocesses).imap_unordered(func, morphologies, chunksize=100)
    errored_paths = list(filter(None, tqdm(results, total=len(morphologies))))
    if errored_paths:
        L.info('Files in error:')
        for path in errored_paths:
            L.info(path)


def fix_non_zero_segments(neuron, zero_length=_ZERO_LENGTH):
    '''Return a neuron with zero length segments removed

    Sections composed of a single zero length segment are deleted, where zero is parametrized
    by zero_length

    Args:
        neuron (str|pathlib.Path|morphio.Morphology|morphio.mut.Morphology): input neuron
        zero_length (float): smallest length of a segment

    Returns:
        a fixed morphio.mut.Morphology
    '''
    neuron = Morphology(neuron)
    to_be_deleted = list()
    for section in neuron.iter():
        points = section.points
        distances = np.linalg.norm(np.diff(points, axis=0), axis=1)
        distances[distances < zero_length] = 0
        indices = np.append(0, np.nonzero(distances)[0] + 1)
        if len(indices) != len(points):
            section.points = section.points[indices]
            section.diameters = section.diameters[indices]
        if len(indices) < 2:
            to_be_deleted.append(section)

    for section in to_be_deleted:
        neuron.delete_section(section)
    return neuron


def annotate_neurolucida(morph_path, checkers=None):
    """Annotate errors on a morphology in neurolucida format.

    Args:
        morph_path (str): absolute path to an ascii morphology
        checkers (dict): dict of checker functons from neurom with function as keys
            and marker data in a dict as values, if None, default checkers are used

    Default checkers include:
        - fat ends
        - z-jumps
        - narrow start
        - dangling branch
        - multifurcation

    Returns:
        annotations to append to .asc file
        dict of error summary
        dict of error markers
    """
    if checkers is None:
        checkers = {
            nc.has_no_fat_ends: {"name": "fat end", "label": "Circle3", "color": "Blue"},
            partial(nc.has_no_jumps, axis="z"): {
                "name": "zjump",
                "label": "Circle2",
                "color": "Green",
            },
            nc.has_no_narrow_start: {"name": "narrow start", "label": "Circle1", "color": "Blue"},
            nc.has_no_dangling_branch: {"name": "dangling", "label": "Circle6", "color": "Magenta"},
            nc.has_multifurcation: {
                "name": "Multifurcation",
                "label": "Circle8",
                "color": "Yellow",
            },
        }

    def _try(checker, neuron):
        """Try to apply a checker, returns True if exception raised, so the checker is bypassed."""
        try:
            return checker(neuron)
        except Exception as e:  # pylint: disable=broad-except
            L.exception("%s failed on %s", checker, morph_path)
            L.exception(e, exc_info=True)
            return CheckResult(True)

    neuron = load_neuron(morph_path)
    results = [_try(checker, neuron) for checker in checkers]
    markers = [
        dict(setting, data=result.info)
        for result, setting in zip(results, checkers.values())
        if not result.status
    ]
    summary = {
        setting["name"]: len(result.info)
        for result, setting in zip(results, checkers.values())
        if result.info
    }
    return annotate(results, checkers.values()), summary, markers


def annotate_neurolucida_all(morph_paths, nprocesses=1):
    """Annotate errors on a list of morphologies in neurolicida format.

    Args:
        morph_paths (list): list of str of paths to morphologies.

    Returns:
        dict annotations to append to .asc file (morph_path as keys)
        dict of dict of error summary (morph_path as keys)
        dict of dict of markers (morph_path as keys)
    """
    summaries, annotations, markers = {}, {}, {}
    for morph_path, result in zip(
        morph_paths, Pool(nprocesses).map(annotate_neurolucida, morph_paths)
    ):
        annotations[str(morph_path)], summaries[str(morph_path)], markers[str(morph_path)] = result
    return annotations, summaries, markers



'''Utils module'''
import json
import logging
from enum import Enum

import numpy as np
from morphio import SectionType
from neurom import NeuriteType, iter_sections

L = logging.getLogger('neuror')


class RepairType(Enum):
    '''The types used for the repair.

    based on
    https://bbpcode.epfl.ch/browse/code/platform/BlueRepairSDK/tree/BlueRepairSDK/src/helper_dendrite.h#n22
    '''
    trunk = 0
    tuft = 1
    oblique = 2
    basal = 3
    axon = 4


def repair_type_map(neuron, apical_section):
    '''Return a dict of extended types'''
    extended_types = dict()
    for section in iter_sections(neuron):
        if section.type == SectionType.apical_dendrite:
            extended_types[section] = RepairType.oblique
        elif section.type == SectionType.basal_dendrite:
            extended_types[section] = RepairType.basal
        elif section.type == SectionType.axon:
            extended_types[section] = RepairType.axon

    if apical_section is not None:
        for section in apical_section.ipreorder():
            extended_types[section] = RepairType.tuft

        # The value for the apical section must be overriden to 'trunk'
        for section in apical_section.iupstream():
            extended_types[section] = RepairType.trunk
    return extended_types


def unit_vector(vector):
    """ Returns the unit vector of the vector.  """
    return vector / np.linalg.norm(vector)


def rotation_matrix(axis, theta):  # pylint: disable=too-many-locals
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    """
    axis = np.asarray(axis)
    axis = axis / np.sqrt(np.dot(axis, axis))
    a = np.cos(theta / 2.0)
    b, c, d = -axis * np.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])


def angle_between(v1, v2):
    """Returns the angle in radians between vectors 'v1' and 'v2'::

    ..code::

        >>> angle_between((1, 0, 0), (0, 1, 0))
        1.5707963267948966
        >>> angle_between((1, 0, 0), (1, 0, 0))
        0.0
        >>> angle_between((1, 0, 0), (-1, 0, 0))
        3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


class RepairJSON(json.JSONEncoder):
    '''JSON encoder that handles numpy types

    In python3, numpy.dtypes don't serialize to correctly, so a custom converter
    is needed.
    '''

    def default(self, o):  # pylint: disable=method-hidden
        if isinstance(o, np.floating):
            return float(o)
        elif isinstance(o, np.integer):
            return int(o)
        elif isinstance(o, np.ndarray):
            return o.tolist()
        elif isinstance(o, NeuriteType):
            return int(o)
        return json.JSONEncoder.default(self, o)


def direction(section):
    '''Return the direction vector of a section

    Args:
        section (morphio.mut.Section): section
    '''
    return np.diff(section.points[[0, -1]], axis=0)[0]


def section_length(section):
    '''Section length

    Args:
        section (morphio.mut.Section): section
    '''
    return np.linalg.norm(direction(section))



'''
The module to run the full repair
'''
import logging
from pathlib import Path

import pandas as pd
from morph_tool.utils import iter_morphology_files

from neuror.main import repair

L = logging.getLogger('neuror')


def repair_all(input_dir, output_dir, seed=0, axons=None, cut_points_dir=None, plots_dir=None):
    '''Repair all morphologies in input folder'''
    for inputfilename in iter_morphology_files(input_dir):
        outfilename = Path(output_dir, inputfilename.name)
        if cut_points_dir:
            cut_points = pd.read_csv(Path(cut_points_dir, inputfilename.name).with_suffix('.csv'))
        else:
            cut_points = None

        if plots_dir is not None:
            name = 'neuron_{}.html'.format(Path(inputfilename).stem.replace(' ', '_'))
            plot_file = str(Path(plots_dir, name))
        else:
            plot_file = None

        try:
            repair(inputfilename, outfilename,
                   seed=seed, axons=axons, cut_leaves_coordinates=cut_points, plot_file=plot_file)
        except Exception as e:  # noqa, pylint: disable=broad-except
            L.warning('%s failed', inputfilename)
            L.warning(e, exc_info=True)



'''The morph-tool command line launcher'''
import json
import shutil
import logging
import os
from pathlib import Path
from pprint import pprint

import click
from morph_tool.utils import iter_morphology_files
from morphio.mut import Morphology  # pylint: disable=import-error
from neurom import load_neuron
from neurom.utils import NeuromJSON

from neuror.cut_plane.detection import CutPlane
from neuror.unravel import DEFAULT_WINDOW_HALF_LENGTH

logging.basicConfig()
L = logging.getLogger('neuror')


@click.group()
@click.option('-v', '--verbose', count=True, default=0,
              help='-v for INFO, -vv for DEBUG')
def cli(verbose):
    '''The CLI entry point'''
    level = (logging.WARNING, logging.INFO, logging.DEBUG)[min(verbose, 2)]
    L.setLevel(level)


@cli.group()
def unravel():
    '''CLI utilities related to unravelling'''


@cli.group()
def cut_plane():
    '''CLI utilities related to cut-plane repair'''


@cli.group()
def sanitize():
    '''CLI utilities related to sanitizing raw morphologies.

    It currently only deals with removing duplicate points but it may do
    more in the future.
    '''


@cut_plane.group()
def compute():
    '''CLI utilities to detect cut planes'''


@cut_plane.group()
def repair():
    '''CLI utilities to repair cut planes'''


@cli.group()
def error_annotation():
    '''CLI utilities related to error annotations'''


@error_annotation.command(short_help='Annotate errors on a morphology')
@click.argument('input_file', type=click.Path(exists=True, file_okay=True))
@click.argument('output_file')
@click.option('--error_summary_file', type=click.Path(file_okay=True), default='error_summary.json',
              help='Path to json file to save error summary')
@click.option('--marker_file', type=click.Path(file_okay=True), default='markers.json',
              help='Path to json file to save markers')
def file(input_file, output_file, error_summary_file, marker_file):
    '''Annotate errors on a morphology.'''
    from neuror.sanitize import annotate_neurolucida

    if Path(input_file).suffix not in ['.asc', '.ASC']:
        raise Exception('Only .asc/.ASC files are allowed, please convert with morph-tool.')

    annotations, summary, markers = annotate_neurolucida(input_file)
    shutil.copy(input_file, output_file)
    with open(output_file, 'a') as morph_file:
        morph_file.write(annotations)
    with open(error_summary_file, 'w') as summary_file:
        json.dump(summary, summary_file, cls=NeuromJSON)
    with open(marker_file, 'w') as m_file:
        json.dump(markers, m_file, cls=NeuromJSON)


@error_annotation.command(short_help='Annotate errors on morphologies')
@click.argument('input_dir')
@click.argument('output_dir', type=click.Path(exists=True, file_okay=False, writable=True))
@click.option('--error_summary_file', type=click.Path(file_okay=True), default='error_summary.json',
              help='Path to json file to save error summary')
@click.option('--marker_file', type=click.Path(file_okay=True), default='markers.json',
              help='Path to json file to save markers')
def folder(input_dir, output_dir, error_summary_file, marker_file):
    '''Annotate errors on a morphologies in a folder.'''
    from neuror.sanitize import annotate_neurolucida_all

    output_dir = Path(output_dir)
    morph_paths = list(iter_morphology_files(input_dir))
    annotations, summaries, markers = annotate_neurolucida_all(morph_paths)
    for morph_path, annotation in annotations.items():
        output_file = output_dir / Path(morph_path).name
        shutil.copy(morph_path, output_file)
        with open(output_file, 'a') as morph_file:
            morph_file.write(annotation)
    with open(error_summary_file, 'w') as summary_file:
        json.dump(summaries, summary_file, indent=4, cls=NeuromJSON)
    with open(marker_file, 'w') as m_file:
        json.dump(markers, m_file, cls=NeuromJSON)


# pylint: disable=function-redefined
@repair.command(short_help='Repair one morphology')
@click.argument('input_file', type=click.Path(exists=True, file_okay=True))
@click.argument('output_file')
@click.option('--plot_file', type=click.Path(file_okay=True), default=None,
              help='Where to save the plot')
@click.option('-a', '--axon-donor', multiple=True,
              help='A morphology that provides a reference axon')
@click.option('--cut-file',
              type=click.Path(exists=True, file_okay=True), default=None,
              help=('Path to a CSV whose columns represents the X, Y and Z '
                    'coordinates of points from which to start the repair'))
def file(input_file, output_file, plot_file, axon_donor, cut_file):
    '''Repair dendrites of a cut neuron'''
    import pandas

    from neuror.main import repair  # pylint: disable=redefined-outer-name

    if cut_file:
        cut_points = pandas.read_csv(Path(cut_file).with_suffix('.csv')).values
    else:
        cut_points = None
    repair(input_file, output_file, axons=axon_donor, cut_leaves_coordinates=cut_points,
           plot_file=plot_file)


# pylint: disable=function-redefined
@repair.command(short_help='Repair all morphologies in a folder')
@click.argument('input_dir')
@click.argument('output_dir', type=click.Path(exists=True, file_okay=False, writable=True))
@click.option('--plot_dir', default=None, type=click.Path(exists=True, file_okay=False,
                                                          writable=True))
@click.option('-a', '--axon-donor', multiple=True,
              help='A morphology that provides a reference axon')
@click.option('--cut-file-dir',
              type=click.Path(exists=True, file_okay=True), default=None,
              help=('A dir with the cut points CSV file for each morphology. '
                    'See also "neuror cut-plane repair file --help".'))
def folder(input_dir, output_dir, plot_dir, axon_donor, cut_file_dir):
    '''Repair dendrites of all neurons in a directory'''
    from neuror.full import repair_all
    repair_all(input_dir, output_dir, axons=axon_donor, cut_points_dir=cut_file_dir,
               plots_dir=plot_dir)


@unravel.command(short_help='Unravel one morphology')
@click.argument('input_file', type=click.Path(exists=True, file_okay=True))
@click.argument('output_file')
@click.option('--mapping-file', type=click.Path(file_okay=True), default=None,
              help=('Path to the file that contains the coordinate mapping before '
                    'and after unravelling'))
@click.option('--window-half-length', default=DEFAULT_WINDOW_HALF_LENGTH)
def file(input_file, output_file, mapping_file, window_half_length):
    '''Unravel a cell'''
    from neuror.unravel import unravel  # pylint: disable=redefined-outer-name
    neuron, mapping = unravel(input_file, window_half_length=window_half_length)
    neuron.write(output_file)
    if mapping_file is not None:
        if not mapping_file.lower().endswith('csv'):
            raise Exception('the mapping file must end with .csv')
        mapping.to_csv(mapping_file)


@unravel.command(short_help='Unravel all morphologies in a folder')
@click.argument('input_dir')
@click.argument('output_dir', type=click.Path(exists=True, file_okay=False, writable=True))
@click.option('--raw-plane-dir', type=click.Path(exists=True, file_okay=False), default=None,
              help='The path to raw cut planes (if None, defaults to INPUT_DIR/planes)')
@click.option('--unravelled-plane-dir', type=click.Path(exists=True, file_okay=False), default=None,
              help='The path to unravelled cut planes (if None, defaults to OUTPUT_DIR/planes)')
@click.option('--window-half-length', default=DEFAULT_WINDOW_HALF_LENGTH)
def folder(input_dir, output_dir, raw_planes_dir, unravelled_planes_dir, window_half_length):
    '''Unravel all cells in a folder'''
    from neuror.unravel import unravel_all
    unravel_all(input_dir, output_dir, raw_planes_dir, unravelled_planes_dir,
                window_half_length=window_half_length)


@cli.command(short_help='Generate PDF with morphology plots')
@click.argument('folders', nargs=-1)
@click.option('--title', '-t', multiple=True)
def report(folders, title):
    '''Generate a PDF with plots of pre and post repair neurons'''
    from neuror.view import view_all
    if not folders:
        print('Need to pass at least one folder')
        return

    if title:
        assert len(title) == len(folders)
    else:
        title = ['Plot {}'.format(i) for i in range(1, len(folders) + 1)]
    view_all(folders, title)


@cli.command(short_help='Fix zero diameters')
@click.argument('input_file')
@click.argument('output_file')
def zero_diameters(input_file, output_file):
    '''Output a morphology where the zero diameters have been removed'''
    from neuror.zero_diameter_fixer import fix_zero_diameters
    neuron = Morphology(input_file)
    fix_zero_diameters(neuron)
    neuron.write(output_file)


# pylint: disable=function-redefined
@sanitize.command(short_help='Sanitize a morphology')
@click.argument('input_file')
@click.argument('output_file')
def file(input_file, output_file):
    '''Sanitize a raw morphology.'''
    from neuror.sanitize import sanitize  # pylint: disable=redefined-outer-name
    sanitize(input_file, output_file)


# pylint: disable=function-redefined
@sanitize.command(short_help='Sanitize all morphologies in a folder')
@click.argument('input_folder')
@click.argument('output_folder')
@click.option('--nprocesses', default=1, help='The number of processes to spawn')
def folder(input_folder, output_folder, nprocesses):
    '''Sanitize all morphologies in the folder.'''
    from neuror.sanitize import sanitize_all
    sanitize_all(input_folder, output_folder, nprocesses=nprocesses)


@cut_plane.group()
def compute():
    '''CLI utilities to compute cut planes'''


def _check_results(result):
    '''Check the result status'''
    if not result:
        L.error('Empty results')
        return -1

    status = result.get('status')
    if status.lower() != 'ok':
        L.warning('Incorrect status: %s', status)
        return 1

    return 0


@compute.command(short_help='Find a 3D cut plane by providing a manual hint')
@click.argument('filename', type=str, required=True)
def hint(filename):
    """Launch the app to manually search for the cut plane. After running the command,
    either click the link in the console or open your browser and go to the address
    shown in the console.

    Example::

       cut-plane compute hint ./tests/data/Neuron_slice.h5
"""
    from neuror.cut_plane.viewer import app, set_neuron
    set_neuron(filename)
    app.run_server(debug=True)


def _export_cut_plane(filename, output, width, display, searched_axes, fix_position):
    '''Find the position of the cut plane (it assumes the plane is aligned along X, Y or Z) for
    morphology FILENAME.

    It returns the cut plane and the positions of all cut terminations.
'''
    if os.path.isdir(filename):
        raise Exception('filename ({}) should not be a directory'.format(filename))

    result = CutPlane.find(filename,
                           width,
                           searched_axes=searched_axes,
                           fix_position=fix_position).to_json()

    if not output:
        pprint(result)
    else:
        with open(output, 'w') as output_file:
            json.dump(result, output_file, cls=NeuromJSON)

    _check_results(result)

    if display:
        from neuror.cut_plane.detection import plot
        plot(load_neuron(filename), result)


@compute.command(short_help='Compute a cut plane for morphology FILENAME')
@click.argument('filename', type=str, required=True)
@click.option('-o', '--output',
              help=('Output name for the JSON file (default=STDOUT)'))
@click.option('-w', '--width', type=float, default=3,
              help='The bin width (in um) of the 1D distributions')
@click.option('-d', '--display', is_flag=True, default=False,
              help='Flag to enable the display control plots')
@click.option('-p', '--plane', type=click.Choice(['x', 'y', 'z']), default=None,
              help='Force the detection along the given plane')
@click.option('--position', type=float, default=None,
              help='Force the position. Requires --plane to be set as well')
def file(filename, output, width, display, plane, position):
    '''Find the position of the cut plane (it assumes the plane is aligned along X, Y or Z) for
    morphology FILENAME.

    It returns the cut plane and the positions of all cut terminations.
    Compute a cut plane and outputs it either as a STDOUT stream or in a file
    if ``-o`` option is passed.
    The control plots can be displayed by passing the ``-d`` option.
    The bin width can be changed with the ``-w`` option (see below)

    Description of the algorithm:

    #. The distribution of all points along X, Y and Z is computed
       and put into 3 histograms.
    #. For each histogram we look at the first and last empty bins
       (that is, the last bin before the histogram starts rising,
       and the first after it reaches zero again). Under the assumption
       that there is no cut plane, the posteriori probability
       of observing this empty bin given the value of the non-empty
       neighbour bin is then computed.
    #. The lowest probability of the 6 probabilities (2 for each axes)
       corresponds to the cut plane.

    .. image:: /_images/distrib_1d.png
       :align: center

    Returns:
       A dictionary with the following items:

       :status: 'ok' if everything went right, else an informative string
       :cut_plane: a tuple (plane, position) where 'plane' is 'X', 'Y' or 'Z'
           and 'position' is the position
       :cut_leaves: an np.array of all termination points in the cut plane
       :figures: if 'display' option was used, a dict where values are tuples (fig, ax)
           for each figure
       :details: A dict currently only containing -LogP of the bin where the cut plane was found

    Example::

       cut-plane compute file -d tests/data/Neuron_slice.h5  -o my-plane.json -w 10
    '''
    _export_cut_plane(filename, output, width, display, plane or ('x', 'y', 'z'), position)


@compute.command(short_help='Compute cut planes for morphologies located in INPUT_DIR')
@click.argument('input_dir')
@click.argument('output_dir', type=click.Path(exists=True, file_okay=False, writable=True))
@click.option('-w', '--width', type=float, default=3,
              help='The bin width (in um) of the 1D distributions')
@click.option('-d', '--display', is_flag=True, default=False,
              help='Flag to enable the display control plots')
@click.option('-p', '--plane', type=click.Choice(['x', 'y', 'z']), default=None,
              help='Force the detection along the given plane')
def folder(input_dir, output_dir, width, display, plane):
    '''Compute cut planes for all morphology in INPUT_DIR and save them into OUTPUT_DIR

    See "cut-plane compute --help" for more information'''
    for inputfilename in iter_morphology_files(input_dir):
        L.info('Seaching cut plane for file: %s', inputfilename)
        outfilename = os.path.join(output_dir, inputfilename.with_suffix('.json'))
        try:
            _export_cut_plane(inputfilename, outfilename, width, display=display,
                              searched_axes=(plane or ('X', 'Y', 'Z')),
                              fix_position=None)
        except Exception as e:  # noqa, pylint: disable=broad-except
            L.warning('Cut plane computation for %s failed', inputfilename)
            L.warning(e, exc_info=True)


@cut_plane.command()
@click.argument('out_filename', nargs=1)
@click.argument('plane_paths', nargs=-1)
def join(out_filename, plane_paths):
    '''Merge cut-planes from json files located at PLANE_PATHS into one.
    The output is writen at OUT_FILENAME

    Example::

       cut-plane join result.json plane1.json plane2.json plane3.json
    '''
    data = []
    for plane in plane_paths:
        with open(plane) as in_f:
            data += json.load(in_f)

    with open(out_filename, 'w') as out_f:
        json.dump(data, out_f)



import json
from collections import defaultdict
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
from mock import patch
from morph_tool.spatial import point_to_section_segment
from morphio import SectionType
from neurom import COLS, NeuriteType, load_neuron
from nose.tools import assert_dict_equal, assert_equal, ok_
from numpy.testing import assert_array_almost_equal, assert_array_equal

import neuror.main as test_module
from neuror.main import Action, Repair, _PARAMS
from neuror.utils import RepairType
from neuror import cut_plane

from .expected_sholl_stats import SHOLL_STATS

DATA_PATH = Path(__file__).parent / 'data'

SIMPLE_PATH = DATA_PATH / 'simple.swc'
SLICE_PATH = DATA_PATH / 'neuron-slice.h5'
SIMPLE = load_neuron(SIMPLE_PATH)
SLICE = load_neuron(SLICE_PATH)


class DummySection:
    def __init__(self, points, children=None):
        self.points = np.array(points)
        self.children = children or []


def test_is_cut_section():
    section = SIMPLE.neurites[0].root_node
    assert_equal(test_module.is_cut_section(section, np.array([[2, 2, 2]])),
                 False)

    assert_equal(test_module.is_cut_section(section, np.array([[0, 0, 0]])),
                 True)


def test_is_branch_intact():
    neurite = SIMPLE.neurites[0]
    assert_equal(test_module.is_branch_intact(neurite.root_node, np.array([[2, 2, 2]])),
                 True)

    assert_equal(test_module.is_branch_intact(neurite.root_node, np.array([[0, 0, 0]])),
                 False)


def test__find_intact_sub_trees():
    obj = Repair(SIMPLE_PATH)
    obj.cut_leaves = np.array([[2, 2, 2]])
    obj._fill_repair_type_map()

    assert_equal(len(obj._find_intact_sub_trees()), 2)

    points = cut_plane.CutPlane.find(SLICE, bin_width=15).cut_leaves_coordinates
    obj = Repair(SLICE_PATH,cut_leaves_coordinates=points)
    obj._fill_repair_type_map()
    intact_sub_trees = obj._find_intact_sub_trees()
    assert_array_equal([section.id for section in intact_sub_trees],
                       [0, 34])

    obj = Repair(DATA_PATH / 'test-cut-apical' / 'simple-apical-cut.swc',
                 cut_leaves_coordinates=[[1.0, 30.0, 0.0]])
    obj._fill_repair_type_map()
    intact_sub_trees = obj._find_intact_sub_trees()
    assert_array_equal([section.id for section in intact_sub_trees],
                       [1, 0, 3])
    assert_array_equal([obj.repair_type_map[section] for section in intact_sub_trees],
                       [RepairType.basal, RepairType.axon, RepairType.tuft])

    filename = DATA_PATH / 'no-intact-basals.h5'
    points = cut_plane.CutPlane.find(filename, bin_width=15).cut_leaves_coordinates
    obj = Repair(filename, cut_leaves_coordinates=points)
    obj._fill_repair_type_map()
    intact_sub_trees = obj._find_intact_sub_trees()
    basals = [section for section in intact_sub_trees
              if section.type == NeuriteType.basal_dendrite]
    assert_equal(len(basals), 78)


def test_section_length():
    assert_equal(test_module.section_length(SIMPLE.neurites[0].root_node), 5)


def test_branching_angles():
    assert_equal(test_module._branching_angles(SIMPLE.neurites[0].root_node),
                 [(0, 1.5707963267948966), (0, 1.5707963267948966)])

    # Test skip too short sections
    assert_equal(test_module._branching_angles(DummySection([[0, 0, 0], [0, 0, 1e-9]])),
                 [])

    # Test skip too short child sections
    tiny_child = DummySection([[0, 0, 0], [0, 0, 1e-9]])
    parent = DummySection([[0, 0, 0], [0, 0, 1]], children=[tiny_child])
    with patch('neuror.main.branch_order'):
        assert_equal(test_module._branching_angles(parent),
                     [])


def test_best_case_angle_data():
    obj = Repair(SIMPLE_PATH)
    obj.info = {'intact_branching_angles': {SectionType.axon: {
        0: [1, 2, 3],
        1: [2, 2],
        2: []
    }}}

    # Use exact data for this specific branching angle
    assert_array_equal(obj._best_case_angle_data(SectionType.axon, 0),
                       [1, 2, 3])

    # No info for this branching angle, use aggregate data
    assert_array_equal(obj._best_case_angle_data(SectionType.axon, 2),
                       [1, 2, 3, 2, 2])


def test_intact_branching_angles():
    obj = Repair(SIMPLE_PATH)
    obj._fill_repair_type_map()
    branches = [neurite.root_node for neurite in obj.neuron.neurites]
    angles = obj._intact_branching_angles(branches)
    assert_array_almost_equal(angles[test_module.RepairType.basal][0],
                              [1.5707963267948966, 1.5707963267948966])


def test__get_sholl_proba():
    sholl_data = {RepairType.axon: {0: {1: {Action.TERMINATION: 2,
                                             Action.BIFURCATION: 2,
                                             Action.CONTINUATION: 4
                                             }}}}

    assert_dict_equal(test_module._get_sholl_proba(sholl_data, RepairType.axon, 0, 1),
                      {Action.TERMINATION: 0.25,
                       Action.BIFURCATION: 0.25,
                       Action.CONTINUATION: 0.5})

    # No info for pseudo_order == 2, re-use data from pseudo_order == 1
    assert_dict_equal(test_module._get_sholl_proba(sholl_data, RepairType.axon, 0, 2),
                      {Action.TERMINATION: 0.25,
                       Action.BIFURCATION: 0.25,
                       Action.CONTINUATION: 0.5})

    # No info for sholl_layer == 1, use default value
    assert_dict_equal(test_module._get_sholl_proba(sholl_data, RepairType.axon, 1, 2),
                      {Action.TERMINATION: 1,
                       Action.BIFURCATION: 0,
                       Action.CONTINUATION: 0})

    # No data at all, use default value
    assert_dict_equal(test_module._get_sholl_proba({RepairType.axon: {1: {}}}, RepairType.axon, 1, 2),
                      {Action.TERMINATION: 1,
                       Action.BIFURCATION: 0,
                       Action.CONTINUATION: 0})


def test_continuation():

    test_module._continuation(
        DummySection([[1.0, 1, 1, 1], [1, 1, 2, 1]]), [0, 0, 0], _PARAMS, 1., 0.1
    )
    test_module._continuation(
        DummySection([[1.0, 1, 1, 1], [1, 1, 2, 1]]), [1, 1, 1], _PARAMS, 1., 0.1
    )


def test_get_origin():
    obj = Repair(SIMPLE)
    section = obj.neuron.section(1)
    obj.repair_type_map = {section: RepairType.basal}
    assert_array_equal(obj._get_origin(section), [0, 0, 0])

    obj.repair_type_map = {section: RepairType.oblique}
    assert_array_equal(obj._get_origin(section), [0, 5, 0])

def test_get_order_offset():
    obj = Repair(SIMPLE_PATH)
    section = obj.neuron.neurites[0].root_node.children[0]
    obj.repair_type_map = {section: RepairType.basal}
    assert_equal(obj._get_order_offset(section), 0)

    obj.repair_type_map = {section: RepairType.oblique}
    assert_equal(obj._get_order_offset(section), 1)


def test__get_sholl_layer():
    section = SIMPLE.neurites[0].root_node
    assert_equal(test_module._get_sholl_layer(
        section, SIMPLE.soma.center, _PARAMS['sholl_layer_size']), 0)


def test_last_segment_vector():
    section = SIMPLE.neurites[0].root_node
    assert_array_equal(test_module._last_segment_vector(section, True),
                       [0, 1, 0])

    assert_array_equal(test_module._last_segment_vector(section, False),
                       [0, 5, 0])


def test__grow_until_sholl_sphere():
    np.random.seed(0)
    neuron = load_neuron(DATA_PATH / 'simple.swc')
    section = neuron.neurites[0].root_node
    test_module._grow_until_sholl_sphere(section, SIMPLE.soma.center, 0, _PARAMS,
                                         lambda diam:diam, 0.1, 1.0)
    assert_array_almost_equal(section.points[:, COLS.XYZ],
                              np.array([[ 0.      ,  0.      ,  0.      ],
                                        [ 0.      ,  5.      ,  0.      ],
                                        [ 0.302276,  9.950119,  0.636358],
                                        [ 1.172574, 14.19536 ,  3.130453]], dtype=np.float32))


def test__compute_sholl_data():
    obj = Repair(SIMPLE_PATH)
    obj._fill_repair_type_map()
    branches = [neurite.root_node for neurite in obj.neuron.neurites]
    sholl_data = obj._compute_sholl_data(branches)
    assert_dict_equal(sholl_data[RepairType.basal][0][0],
                      {Action.TERMINATION: 0, Action.CONTINUATION: 0, Action.BIFURCATION: 1})
    assert_dict_equal(sholl_data[RepairType.basal][0][1],
                      {Action.TERMINATION: 2, Action.CONTINUATION: 0, Action.BIFURCATION: 0})
    assert_dict_equal(sholl_data[RepairType.oblique], {})
    assert_dict_equal(sholl_data[RepairType.axon],
                      {0: {0: {Action.BIFURCATION: 1,
                               Action.CONTINUATION: 0,
                               Action.TERMINATION: 0},
                           1: {Action.BIFURCATION: 0,
                               Action.CONTINUATION: 0,
                               Action.TERMINATION: 2}}})


def test__external_apical_point():
    obj = Repair(DATA_PATH / 'repair_type.asc', apical_point=[1.0, 0.0, 0.0])
    assert_array_almost_equal(obj.apical_section.points,
                              np.array([[0., 0., 4., 1.],
                                        [0., 0., 0., 0.5],
                                        [1., 0., 0., 0.5]], dtype=np.float32))


def test__fill_repair_type_map():
    obj = Repair(DATA_PATH / 'repair_type.asc')
    obj.apical_section = obj.neuron.sections[3]
    # Pre check to be sure that the apical point is the one it should be
    assert_array_almost_equal(obj.apical_section.points,
                              np.array([[0., 0., 4., 1.],
                                        [0., 0., 0., 0.5],
                                        [1., 0., 0., 0.5]], dtype=np.float32))

    obj._fill_repair_type_map()
    assert_dict_equal({sec.id: v for sec, v in obj.repair_type_map.items()},
                      {
                          0: test_module.RepairType.axon,
                          1: test_module.RepairType.trunk,
                          2: test_module.RepairType.oblique,
                          3: test_module.RepairType.trunk,
                          4: test_module.RepairType.tuft,
                          5: test_module.RepairType.tuft,
                          6: test_module.RepairType.basal,
    })


def json_compatible_dict(dict_):
    '''Remap the dict keys so that it can be saved to JSON'''
    if not isinstance(dict_, dict):
        return dict_

    result = dict()
    for k, v in dict_.items():
        if isinstance(k, (SectionType, Action, int)):
            k = str(k)
        result[k] = json_compatible_dict(v)

    return result

def test__compute_statistics_for_intact_subtrees():
    input_file = DATA_PATH / 'neuron-slice.h5'
    points = cut_plane.CutPlane.find(input_file, bin_width=15).cut_leaves_coordinates
    obj = Repair(input_file, cut_leaves_coordinates=points)
    obj._fill_repair_type_map()
    obj._fill_statistics_for_intact_subtrees()

    with open(DATA_PATH / 'neuron-slice-sholl-data.json') as f:
        expected = json.load(f)

    basal_data = expected['SectionType.basal_dendrite']

    actual = json_compatible_dict(obj.info['sholl'][RepairType.basal])
    for layer in basal_data:
        for order in basal_data[layer]:
            for action in basal_data[layer][order]:
                assert_equal(basal_data[layer][order][action],
                             actual[layer][order][action])

def test__grow():
    obj = Repair(SIMPLE_PATH, seed=1)
    leaf = obj.neuron.sections[2]
    obj.repair_type_map = {section: RepairType.basal for section in obj.neuron.sections}
    obj.info = {
        'sholl': {RepairType.basal: {0: {1: {Action.BIFURCATION: 0.5, Action.TERMINATION: 0, Action.CONTINUATION: 0.5}}}},
        'dendritic_sections': obj.neuron.sections,
        'intact_branching_angles': {RepairType.basal: {1: [0.2]}}
    }
    obj.current_trunk_radius = 1.0
    obj._grow(leaf, 0, obj.neuron.soma.center)
    assert_equal(len(obj.neuron.sections), 8)


def test_repair_axon():
    filename = DATA_PATH / 'real-with-axon.asc'
    with TemporaryDirectory('test-cli-axon') as tmp_folder:
        outfilename = Path(tmp_folder, 'out.asc')
        test_module.repair(filename, outfilename, axons=[filename])
        neuron_in = load_neuron(filename)
        neuron_out = load_neuron(outfilename)
        axon = neuron_out.section(41)
        ok_(axon.type == NeuriteType.axon)
        assert_array_equal(neuron_in.section(40).points[0],
                           neuron_out.section(40).points[0])
        ok_(len(neuron_out.section(41).points) > len(neuron_in.section(41).points))

        # Test disactivating the axon repair
        repair_flags = {RepairType.axon: False}
        test_module.repair(filename, outfilename, axons=[filename], repair_flags=repair_flags)
        neuron_out = load_neuron(outfilename)
        axon = neuron_out.section(41)
        ok_(axon.type == NeuriteType.axon)
        assert_array_equal(neuron_in.section(41).points[0],
                           neuron_out.section(41).points[0])

        ok_(len(neuron_out.section(41).points) == len(neuron_in.section(41).points),
            'The section should not have been regrown')


def test_repair_no_intact_axon():
    filename = DATA_PATH / 'no-intact-basals.h5'
    with TemporaryDirectory('test-cli-axon') as tmp_folder:
        outfilename = Path(tmp_folder, 'out.asc')
        test_module.repair(filename, outfilename, axons=[filename])


def test_repair_no_trunk():
    '''Test repair when the morph has oblique sections but no trunk'''
    filename = DATA_PATH / 'Fluo42_right.h5'
    with TemporaryDirectory('test-no-trunk') as tmp_folder:
        outfilename = Path(tmp_folder, 'out.asc')
        test_module.repair(filename, outfilename, legacy_detection=True)


def test_legacy_compare_with_legacy_result():
    '''Comparing results with the old repair launch with the following commands:

    repair --dounravel 0 --inputdir /gpfs/bbp.cscs.ch/project/proj83/home/gevaert/morph-release/morph_release_old_code-2020-07-27/output/04_ZeroDiameterFix --input rp120430_P-2_idA --overlap=true --incremental=false --restrict=true --distmethod=mirror

    The arguments are the one used in the legacy morphology workflow.
    '''
    neuron = load_neuron(DATA_PATH / 'compare-bbpsdk/rp120430_P-2_idA.h5')
    obj = test_module.Repair(inputfile=DATA_PATH / 'compare-bbpsdk/rp120430_P-2_idA.h5', legacy_detection=True)

    cut_sections = {point_to_section_segment(neuron, point)[0]
                    for point in obj.cut_leaves}

    legacy_cut_sections = {
        13,14,17,18,38,39,40,45,58,67,68,69,73,75,76,93,94,101,102,103,105,106,
        109,110,111,120,124,125,148,149,150,156,157,158,162,163,164,166,167,168,
        169,192,201,202,203,205,206,208
    }
    assert_equal(cut_sections, legacy_cut_sections)

    obj._fill_repair_type_map()

    types = defaultdict(list)
    for k, v in obj.repair_type_map.items():
        types[v].append(k)


    # offset due to the first section id in the old soft being the soma
    offset = 1

    # These numbers come from the attribute 'apical' from the h5py group 'neuron1'
    section_id, segment_id = 134, 8


    assert_equal(obj.apical_section.id + offset, section_id)

    assert_equal(len(obj.apical_section.points) - 1, segment_id)

    assert_array_equal([section.id + offset for section in types[RepairType.basal]],
                       [90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106,
                        107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121,
                        122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132])

    assert_array_equal([0] + [section.id + offset for section in types[RepairType.axon]],
                       [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
                        21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39,
                        40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58,
                        59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77,
                        78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89])
    assert_array_equal([section.id + offset for section in types[RepairType.oblique]],
                       [217, 218, 219])
    assert_array_equal([section.id + offset for section in types[RepairType.trunk]],
                       [133, 134])

    expected_tufts = {135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149,
                      150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164,
                      165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179,
                      180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194,
                      195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209,
                      210, 211, 212, 213, 214, 215, 216}
    actual_tufts = {section.id + offset for section in types[RepairType.tuft]}
    assert_equal(actual_tufts, expected_tufts)


    expected_axons = {1, 2, 77, 3, 70, 4, 59, 5, 46, 6, 41, 7, 40, 8, 39, 9, 38, 10, 19, 11, 16, 12,
                      15, 13, 14, 17, 18, 20, 37, 21, 34, 22, 31, 23, 28, 24, 27, 25, 26, 29, 30,
                      32, 33, 35, 36, 42, 45, 43, 44, 47, 58, 48, 53, 49, 52, 50, 51, 54, 55, 56,
                      57, 60, 69, 61, 66, 62, 63, 64, 65, 67, 68, 71, 76, 72, 75, 73, 74, 78, 83,
                      79, 82, 80, 81, 84, 85, 86, 89, 87, 88}
    actual_axons = {section.id + offset for section in types[RepairType.axon]}
    assert_equal(actual_axons, expected_axons)


    intacts = defaultdict(list)

    for sec in obj._find_intact_sub_trees():
        intacts[obj.repair_type_map[sec]].append(sec)

    assert_equal([sec.id + offset for sec in intacts[RepairType.trunk]],
                 [])
    assert_equal([sec.id + offset for sec in intacts[RepairType.oblique]],
                 [217])
    assert_equal({sec.id + offset for sec in intacts[RepairType.tuft]},
                 {135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 151, 152, 153,
                  154, 155, 159, 160, 161, 165, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179,
                  180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 193, 194, 195, 196,
                  197, 198, 199, 200, 204, 207, 209, 210, 211, 212, 213, 214, 215, 216})


def test_legacy_sholl_data():
    '''Comparing results with the old repair launch with the following commands:

    repair --dounravel 0 --inputdir /gpfs/bbp.cscs.ch/project/proj83/home/gevaert/morph-release/morph_release_old_code-2020-07-27/output/04_ZeroDiameterFix --input rp120430_P-2_idA --overlap=true --incremental=false --restrict=true --distmethod=mirror

    The arguments are the one used in the legacy morphology workflow.
    '''
    neuron = load_neuron(DATA_PATH / 'compare-bbpsdk/rp120430_P-2_idA.h5')
    obj = test_module.Repair(inputfile=DATA_PATH / 'compare-bbpsdk/rp120430_P-2_idA.h5', legacy_detection=True)
    obj._fill_repair_type_map()
    obj._fill_statistics_for_intact_subtrees()

    key_mapping = {
        'basal': RepairType.basal,
        'oblique': RepairType.oblique,
        'tuft': RepairType.tuft
    }
    expected_stats = {key_mapping[k]: v for k,v in SHOLL_STATS.items()}

    flattened = {(i, j): [obj.info['sholl'][RepairType.oblique][i][j][action] for action in
                          [Action.BIFURCATION, Action.CONTINUATION, Action.TERMINATION]]
                 for i in obj.info['sholl'][RepairType.oblique].keys()
                 for j in obj.info['sholl'][RepairType.oblique][i].keys()}

    assert_dict_equal(flattened, expected_stats[RepairType.oblique])

def test_repair_only_axon():
    '''Test repairing a morph which has only an axon (but no dendrites) does not crash'''
    test_module.Repair(DATA_PATH / 'just-axon.asc')
    ok_(True)

def test_legacy_compare_with_legacy_result2():
    '''Comparing results with the old repair launch with the following commands:

    repair --dounravel 0 --inputdir /gpfs/bbp.cscs.ch/project/proj83/home/gevaert/morph-release/morph_release_old_code-2020-07-27/output/04_ZeroDiameterFix --input vd100714B_idB --overlap=true --incremental=false --restrict=true --distmethod=mirror

    The arguments are the one used in the legacy morphology workflow.
    '''
    neuron = load_neuron(DATA_PATH / 'compare-bbpsdk/vd100714B_idB.h5')
    obj = test_module.Repair(inputfile=DATA_PATH / 'compare-bbpsdk/vd100714B_idB.h5', legacy_detection=True)

    cut_sections = {point_to_section_segment(neuron, point)[0]
                    for point in obj.cut_leaves}

    legacy_cut_sections = {62,64,65,69,73,77,78,85,87,88,89,91,93,94,115,116,119,120,125,126,130,
                           133,136,137,138,140,142,144,145,147,150,151,152,159,165,171,172,175,177
                           ,179,180,182,184,188,191,200,202,204,205,207,208,209,211,215,217,218,219,
                           220,238,239,241,247,248,250,251,252,253,256,257,258,261,262,264,266,267,
                           283,288,289,290,291,293,294,295,316,318,320,322,324,326,328,330,331,337,
                           338,339,340,343,344,345,351,357,359,362,363,371,372,375,377,378,384,385,
                           386,387,388,390,391,394,416,426,427,429,430,431,438,439,440,441,453,466,
                           468,470,471,481,486,487,488,489,527,528,529,533,534,538,540,541,543,545,
                           548,549,551,572,573,574,576,577,581,583,584,588,595,596,598,599,602,607,
                           608,609,610,613,614,615,617,620,622,623,624,626,637,639,640,645,647,648,
                           649,650,653,654,665,666,667,670,677,678,679,680,689,691,693,694,703,716,
                           717,721,723,725,726}
    assert_equal(cut_sections, legacy_cut_sections)

    obj._fill_repair_type_map()

    types = defaultdict(list)
    for k, v in obj.repair_type_map.items():
        types[v].append(k)


    # offset due to the first section id in the old soft being the soma
    offset = 1

    assert_equal(obj.apical_section, None)

    assert_equal({section.id + offset for section in types[RepairType.basal]},
                 {650, 651, 668, 671, 702, 719, 652, 655, 653, 654, 656, 663, 657, 662, 658, 661,
                  659, 660, 664, 667, 665, 666, 669, 670, 672, 681, 673, 680, 674, 679, 675, 678,
                  676, 677, 682, 695, 683, 684, 685, 686, 687, 690, 688, 689, 691, 692, 693, 694,
                  696, 699, 697, 698, 700, 701, 703, 704, 705, 706, 707, 710, 708, 709, 711, 718,
                  712, 713, 714, 715, 716, 717, 720, 727, 721, 722, 723, 724, 725, 726, 728, 735,
                  729, 730, 731, 734, 732, 733, 736, 737, 738})

    assert_array_equal([section.id + offset for section in types[RepairType.oblique]],
                       [])
    assert_array_equal([section.id + offset for section in types[RepairType.trunk]],
                       [])

    assert_equal({section.id + offset for section in types[RepairType.tuft]},
                 set())

    intacts = defaultdict(list)

    for sec in obj._find_intact_sub_trees():
        intacts[obj.repair_type_map[sec]].append(sec)

    # Since there is no apical dendrite, all of those are empty
    for extended_type in [RepairType.trunk, RepairType.oblique, RepairType.tuft]:
        assert_equal(intacts[extended_type], [])

    assert_equal({sec.id + offset for sec in intacts[RepairType.basal]},
                 {651, 668, 671, 702, 719, 652, 655, 656, 663, 657, 662, 658, 661, 659, 660, 664,
                  669, 672, 681, 673, 674, 675, 676, 682, 695, 683, 684, 685, 686, 687, 690, 688,
                  692, 696, 699, 697, 698, 700, 701, 704, 705, 706, 707, 710, 708, 709, 711, 718,
                  712, 713, 714, 715, 720, 727, 722, 724, 728, 735, 729, 730, 731, 734, 732, 733,
                  736, 737, 738})



import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

from click.testing import CliRunner
from nose.tools import assert_equal

from neuror.cli import cli

DATA = Path(__file__).parent / 'data'


def test_error_annotation_file():
    runner = CliRunner()
    with TemporaryDirectory('test-file') as folder:
        result = runner.invoke(cli, ['error-annotation', 'file',
                                     str(DATA / 'test-error-detection/error-morph.asc'),
                                     str(Path(folder, 'out.asc'))])
        assert_equal(result.exit_code, 0, result.exception)


def test_error_annotation_folder():
    runner = CliRunner()
    with TemporaryDirectory('test-cli-folder') as folder:
        result = runner.invoke(cli, ['error-annotation', 'folder',
                                     str(DATA / 'test-error-detection'),
                                     folder])
        assert_equal(result.exit_code, 0, result.exception)
        assert_equal(set(str(path.relative_to(folder)) for path in Path(folder).rglob('*')),
                     {'simple.asc', 'error-morph.asc'})


def test_repair_file():
    runner = CliRunner()
    with TemporaryDirectory('test-file') as folder:
        result = runner.invoke(cli, ['cut-plane', 'repair', 'file',
                                     str(DATA / 'real.asc'),
                                     str(Path(folder, 'out.asc'))])
        assert_equal(result.exit_code, 0, result.exception)


def test_repair_folder():
    runner = CliRunner()
    with TemporaryDirectory('test-cli-folder') as folder:
        result = runner.invoke(cli, ['cut-plane', 'repair', 'folder',
                                     str(DATA / 'input-repair-all'),
                                     folder])
        assert_equal(result.exit_code, 0, result.exception)
        assert_equal(set(str(path.relative_to(folder)) for path in Path(folder).rglob('*')),
                     {'simple.asc', 'simple2.asc'})


def test_repair_with_plane():
    runner = CliRunner()
    input_path = DATA / 'input-repair-all'
    with TemporaryDirectory('test-cli-folder') as folder:
        result = runner.invoke(cli, ['cut-plane', 'repair', 'folder',
                                     str(input_path),
                                     folder,
                                     '--cut-file-dir', str(input_path / 'planes')])
        assert_equal(result.exit_code, 0, result.exc_info)


def test_cli_axon():
    runner = CliRunner()
    with TemporaryDirectory('test-cli-axon') as tmp_folder:
        tmp_folder = Path(tmp_folder)
        result = runner.invoke(cli, ['cut-plane', 'repair', 'file',
                                     '-a', str(DATA / 'real-with-axon.asc'),
                                     str(DATA / 'real-with-axon.asc'),
                                     str(tmp_folder / 'output.asc')])
        assert_equal(result.exit_code, 0)


def test_sanitize():
    runner = CliRunner()
    with TemporaryDirectory('test-cli-axon') as tmp_folder:
        tmp_folder = Path(tmp_folder)
        result = runner.invoke(cli, ['sanitize', 'file',
                                     str(DATA / 'simple-with-duplicates.asc'),
                                     str(tmp_folder / 'output.asc')])
        assert_equal(result.exit_code, 0)

        result = runner.invoke(cli, ['sanitize', 'folder',
                                    str(DATA), str(tmp_folder)])
        assert_equal(result.exit_code, 0)



from pathlib import Path

from morph_tool import apical_point_section_segment
from morph_tool.spatial import point_to_section_segment
from neurom import load_neuron
from nose.tools import assert_equal, assert_raises
from numpy.testing import assert_array_almost_equal, assert_array_equal

import neuror.cut_plane.legacy_detection as test_module
from neuror.utils import repair_type_map

DATA_PATH = Path(__file__).parent / 'data'

SIMPLE = load_neuron(DATA_PATH / 'simple.swc')

# OFFSET to go from BBPSDK section IDs to neurom ones
OFFSET = 1

def test_legacy():
    points, sign = test_module.internal_cut_detection(SIMPLE, 'y')
    assert_array_equal(points,
                       [[ 6., -4., 0.],
                        [-5., -4., 0.]])
    assert_equal(sign, -1)

    assert_raises(Exception, test_module.cut_detect, SIMPLE, 'z')

def test_get_obliques():
    '''Comparing results with the old repair launch with the following commands:

    repair --dounravel 0 --inputdir /gpfs/bbp.cscs.ch/project/proj83/home/bcoste/release/out-new/01_ConvertMorphologies --input rp120430_P-2_idA --overlap=true --incremental=false --restrict=true --distmethod=mirror

    The arguments are the one used in the legacy morphology workflow.
    '''
    neuron = load_neuron(DATA_PATH / 'compare-bbpsdk/rp120430_P-2_idA.h5')

    apical_section = neuron.sections[apical_point_section_segment(neuron)[0]]
    extended_types = repair_type_map(neuron, apical_section)
    assert_equal([sec.id + OFFSET for sec in test_module.get_obliques(neuron, extended_types)],
                 [217])


def test_children_ids():
    apical_section_id = 133
    neuron = load_neuron(DATA_PATH / 'compare-bbpsdk/rp120430_P-2_idA.h5')
    section = neuron.sections[apical_section_id]
    assert_equal([sec.id + OFFSET for sec in test_module.children_ids(section)],
                 [134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, ])

def test_legacy_compare_with_legacy_result():
    '''Comparing results with the old repair launch with the following commands:

    repair --dounravel 0 --inputdir /gpfs/bbp.cscs.ch/project/proj83/home/bcoste/release/out-new/01_ConvertMorphologies --input rp120430_P-2_idA --overlap=true --incremental=false --restrict=true --distmethod=mirror

    The arguments are the one used in the legacy morphology workflow.
    '''
    neuron = load_neuron(DATA_PATH / 'compare-bbpsdk/rp120430_P-2_idA.h5')
    points, sign = test_module.internal_cut_detection(neuron, 'z')
    assert_equal(sign, 1)
    cut_sections = {point_to_section_segment(neuron, point)[0]
                    for point in points}

    legacy_cut_sections = {13,14,17,18,38,39,40,45,58,67,68,69,73,75,76,93,94,101,102,103,105,106,109,110,111,120,124,125,148,149,150,156,157,158,162,163,164,166,167,168,169,192,201,202,203,205,206,208}
    assert_equal(cut_sections, legacy_cut_sections)
